import logging
import os
import xml.etree.ElementTree as ET
from typing import Dict, List, Optional, Tuple

from ..gpx_elements import Gpx
from .writer import Writer

DEFAULT_NORMAL_STYLE = {"color": "ff0000ff", "width": 2, "fill": 0}

DEFAULT_HIGHLIGHT_STYLE = {"color": "ff0000ff", "width": 2, "fill": 0}

DEFAULT_STYLES = [
    ("normal", DEFAULT_NORMAL_STYLE),
    ("highlight", DEFAULT_HIGHLIGHT_STYLE),
]


class KMLWriter(Writer):
    """
    KML file writer.
    """

    def __init__(
        self,
        gpx: Gpx = None,
        properties: bool = True,
        metadata: bool = True,  # unused
        way_points: bool = True,
        routes: bool = True,
        extensions: bool = True,
        ele: bool = True,
        time: bool = True,
        precisions: Dict = None,
        time_format: str = None,
        styles: List[Tuple[str, Dict]] = None,
    ) -> None:
        """
        Initialize GPXWriter instance.
        """
        super().__init__(gpx, precisions, time_format)
        self.file_name: str = ""
        self.kml_string: str = ""

        # Parameters
        self.properties: bool = properties
        self.metadata: bool = metadata
        self.way_points: bool = way_points
        self.routes: bool = routes
        self.extensions: bool = extensions
        self.ele: bool = ele
        self.time: bool = time

        self.styles = styles if styles is not None else DEFAULT_STYLES

        # Utility attributes
        self.gpx_string: str = ""
        self.kml_root = None

    def add_pair(self, element: ET.Element, key: str, style_url: str) -> ET.Element:
        """
        Add StyleMap to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.
        key : str
            Key.
        style_url : str
            Style URL.

        Returns
        -------
        ET.Element
            KML element.
        """
        pair_ = ET.SubElement(element, "Pair")
        pair_, _ = self.add_subelement(pair_, "key", key)
        pair_, _ = self.add_subelement(pair_, "styleUrl", style_url)
        return element

    def add_stylemap(self, element: ET.Element, id_: str) -> ET.Element:
        """
        Add StyleMap to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.
        id_ : str
            StyleMap element id.

        Returns
        -------
        ET.Element
            KML element.
        """
        stylemap_ = ET.SubElement(element, "StyleMap")
        self.setIfNotNone(stylemap_, "id", id_)
        style_id = 1
        for style_key, _ in self.styles:  # _ used to be called style
            stylemap_ = self.add_pair(stylemap_, style_key, "#style" + str(style_id))
            style_id += 1
        return element

    def add_polystyle(self, element: ET.Element, style: Dict) -> ET.Element:
        """
        Add PolyStyle to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.
        style : Dict
            Style.

        Returns
        -------
        ET.Element
            KML element.
        """
        polystyle_ = ET.SubElement(element, "PolyStyle")
        try:
            polystyle_, _ = self.add_subelement_number(
                polystyle_, "fill", style["fill"]
            )
        except:
            logging.warning("No fill attribute in style")
            polystyle_, _ = self.add_subelement_number(polystyle_, "fill", 0)
        return element

    def add_linestyle(self, element: ET.Element, style: Dict) -> ET.Element:
        """
        Add LineStyle to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.
        style : Dict
            Style.

        Returns
        -------
        ET.Element
            KML element.
        """
        linestyle_ = ET.SubElement(element, "LineStyle")
        try:
            linestyle_, _ = self.add_subelement(linestyle_, "color", style["color"])
        except:
            logging.warning("No color attribute in style")
            linestyle_, _ = self.add_subelement(linestyle_, "color", "ff0000ff")
        try:
            linestyle_, _ = self.add_subelement_number(
                linestyle_, "width", style["width"]
            )
        except:
            logging.warning("No width attribute in style")
            linestyle_, _ = self.add_subelement_number(linestyle_, "width", 2)
        return element

    def add_style(self, element: ET.Element, id_: str, style: Dict) -> ET.Element:
        """
        Add Style to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.
        id_ : str
            Style element id.
        style : Dict
            Line style.

        Returns
        -------
        ET.Element
            KML element.
        """
        style_ = ET.SubElement(element, "Style")
        self.setIfNotNone(style_, "id", id_)
        style_ = self.add_linestyle(style_, style)
        style_ = self.add_polystyle(style_, style)
        return element

    def add_linestring(self, element: ET.Element) -> ET.Element:
        """
        Add LineString to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.

        Returns
        -------
        ET.Element
            KML element.
        """
        linestring_ = ET.SubElement(element, "LineString")
        linestring_, _ = self.add_subelement_number(linestring_, "tessellate", 1)
        if self.ele:
            coordinates = self.gpx.to_csv(
                path=None, values=["lon", "lat", "ele"], header=False
            ).replace("\n", " ")
        else:
            coordinates = self.gpx.to_csv(
                path=None, values=["lon", "lat"], header=False
            ).replace("\n", " ")
        linestring_, _ = self.add_subelement(linestring_, "coordinates", coordinates)
        return element

    def add_placemark(self, element: ET.Element) -> ET.Element:
        """
        Add Placemark to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.

        Returns
        -------
        ET.Element
            KML element.
        """
        placemark_ = ET.SubElement(element, "Placemark")
        placemark_, _ = self.add_subelement(placemark_, "name", self.gpx.name())
        placemark_, _ = self.add_subelement(placemark_, "styleUrl", "#stylemap")
        placemark_ = self.add_linestring(placemark_)
        return element

    def add_document(self, element: ET.Element) -> ET.Element:
        """
        Add Document to KML element.

        Parameters
        ----------
        element : ET.Element
            KML element.

        Returns
        -------
        ET.Element
            KML element.
        """
        document_ = ET.SubElement(element, "Document")
        document_, _ = self.add_subelement(document_, "name", self.file_name)
        id_ = 1
        for _, style in self.styles:  # _ used to be called style_key
            document_ = self.add_style(document_, "style" + str(id), style)
            id_ += 1
        document_ = self.add_stylemap(document_, "stylemap")
        document_ = self.add_placemark(document_)
        return element

    def add_root_document(self) -> None:
        """
        Add Document element to the KML root element.
        """
        logging.info("Preparing Document...")

        self.kml_root = self.add_document(self.kml_root)

    def add_root_properties(self) -> None:
        """
        Add properties to the GPX root element.
        """
        logging.info("Preparing properties...")

        # According to .kml file from Google Earth Pro
        self.kml_root.set("xmlns", "http://www.opengis.net/kml/2.2")
        self.kml_root.set("xmlns:gx", "http://www.google.com/kml/ext/2.2")
        self.kml_root.set("xmlns:kml", "http://www.opengis.net/kml/2.2")
        self.kml_root.set("xmlns:atom", "http://www.w3.org/2005/Atom")

    def gpx_to_string(self) -> str:
        """
        Convert Gpx instance to a string (the content of a .kml file).

        Returns:
            str: String corresponding to the Gpx instance.
        """
        if self.gpx is not None:
            logging.info("Start convertion from GPX to string")
            # Reset string
            self.kml_string = ""

            # Root
            self.kml_root = ET.Element("kml")

            # Properties
            if self.properties:
                self.add_root_properties()

            # Document
            self.add_root_document()

            # Convert data to string
            logging.info("Converting GPX to string...")
            self.gpx_string = ET.tostring(self.kml_root, encoding="unicode")
            # self.gpx_string = ET.tostring(kml_root, encoding="utf-8")

            logging.info("GPX successfully converted to string:\n%s", self.gpx_string)

            return self.gpx_string

    def write_gpx(self):
        """
        Convert Gpx instance to string and write to .kml file.
        """
        # Open/create KML file
        try:
            f = open(self.file_path, "w", encoding="utf-8")
        except OSError:
            logging.exception("Could not open/read file: %s", self.file_path)
            raise
        # Write KML file
        with f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>')
            f.write(self.gpx_string)

    def write(
        self,
        file_path: str,
        styles: Optional[List[Tuple[str, Dict]]] = None,
        xml_schema: bool = False,
        xml_extensions_schemas: bool = False,
    ) -> bool:
        """
        Handle writing.

        Parameters
        ----------
        file_path : str
            Path to write the KML file.
        styles : Optional[List[Tuple[str, Dict]]], optional
            List of (style_id, style) tuples, by default None
        check_xml_schemas : bool, optional
            Toggle schema verification after writting, by default False

        Returns
        -------
        bool
            bool: Return False if written file does not follow checked schemas. Return True otherwise.
        """
        # Handle path
        directory_path = os.path.dirname(os.path.realpath(file_path))
        if not os.path.exists(directory_path):
            logging.error("Provided path does not exist")
            return False
        self.file_path = file_path
        self.file_name = os.path.basename(self.file_path)

        # Update style
        if styles is not None:
            self.styles = styles

        # Write .kml file
        self.gpx_to_string()
        self.write_gpx()

        # Check XML schemas
        res = True
        if xml_schema or xml_extensions_schemas:
            res = self.check_xml_schemas(xml_schema, xml_extensions_schemas)

        return res



from .gpx_element import GpxElement


class Email(GpxElement):
    """
    emailType element in GPX file.
    """

    fields = ["id", "domain"]
    mandatory_fields = ["id", "domain"]

    def __init__(self, tag: str = "email", id: str = None, domain: str = None) -> None:
        """
        Initialize Email instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "email"
        id : str, optional
            Identifier, by default None
        domain : str, optional
            Domain, by default None
        """
        self.tag: str = tag
        self.id: str = id
        self.domain: str = domain



from .gpx_element import GpxElement


class Link(GpxElement):
    """
    linkType element in GPX file.
    """

    fields = ["href", "text", "type"]
    mandatory_fields = ["href"]

    def __init__(
        self, tag: str = "link", href: str = None, text: str = None, type_: str = None
    ) -> None:
        """
        Initialize Link instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "link"
        href : str, optional
            Hyper reference, by default None
        text : str, optional
            Text, by default None
        type : str, optional
            Type, by default None
        """
        self.tag: str = tag
        self.href: str = href
        self.text: str = text
        self.type: str = type_



import logging
import os
from datetime import datetime
from typing import List, Optional

from fitparse import FitFile

from ..gpx_elements import Gpx, Track, TrackSegment, WayPoint
from .parser import DEFAULT_PRECISION, Parser


class FitParser(Parser):
    """
    Fit file parser.
    """

    def __init__(self, file_path: Optional[str] = None) -> None:
        """
        Initialize FitParser instance.

        Args:
            file_path (str, optional): Path to the file to parse. Defaults to None.
        """
        if not (file_path.endswith(".fit") or file_path.endswith(".FIT")):
            return
        super().__init__(file_path)

        if self.file_path is not None and os.path.exists(self.file_path):
            self.parse()
        else:
            logging.warning("File path does not exist")

    def _set_time_format(self, time):
        """
        Set the time format used in FIT file.
        """
        if time is None:
            logging.warning("No time element in FIT file.")
            return

        try:
            datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%fZ")
            self.time_format = "%Y-%m-%dT%H:%M:%S.%fZ"
        except:
            self.time_format = "%Y-%m-%dT%H:%M:%SZ"  # default time format

    def _semicircles_to_deg(self, l: List) -> List:
        """
        Convert semicircle data from FIT file to dms data.

        Args:
            l (list): List of semicircle values.

        Returns:
            list: List of dms values.
        """
        const = 180 / 2**31
        l = map(lambda x: const * x, l)
        return l

    def _parse(self):
        """
        Parse FIT file and store data in a Gpx element.
        """
        lat_data = []
        lon_data = []
        alt_data = []
        time_data = []

        units = {"alt": "", "lat": "", "lon": ""}

        fit_file = FitFile(self.file_path)

        for record in fit_file.get_messages("record"):
            for record_data in record:
                if record_data.name == "position_lat":
                    lat_data.append(record_data.value)
                    if units["lat"] == "":
                        units["lat"] = record_data.units
                        # Temporary (may change if units["lat"] == "semicircles")
                        self.precisions["lat_lon"] = self.find_precision(
                            record_data.value
                        )
                if record_data.name == "position_long":
                    lon_data.append(record_data.value)
                    if units["lon"] == "":
                        units["lon"] = record_data.units
                if record_data.name == "altitude":
                    alt_data.append(record_data.value)
                    if units["alt"] == "":
                        units["alt"] = record_data.units
                        self.precisions["elevation"] = self.find_precision(
                            record_data.value
                        )
                if record_data.name == "timestamp":
                    time_data.append(record_data.value)
                    self._set_time_format(record_data.value)

        # Convert semicircles data to radians ??
        if units["lat"] == "semicircles":
            self.precisions["lat_lon"] = DEFAULT_PRECISION
            lat_data = self._semicircles_to_deg(lat_data)
        if units["lon"] == "semicircles":
            lon_data = self._semicircles_to_deg(lon_data)

        # Store FIT data in Gpx element
        trkpt = []
        for lat, lon, alt, time in list(zip(lat_data, lon_data, alt_data, time_data)):
            trkpt.append(WayPoint("trkpt", lat, lon, alt, time))
        trkseg = TrackSegment(trkpt=trkpt)
        trk = Track(trkseg=[trkseg])
        self.gpx.trk = [trk]

    def _add_properties(self):
        self.gpx.creator = "ezGPX"
        self.gpx.xmlns = "http://www.topografix.com/GPX/1/1"
        self.gpx.version = "1.1"
        self.gpx.xmlns_xsi = "http://www.w3.org/2001/XMLSchema-instance"
        self.gpx.xsi_schema_location = [
            "http://www.topografix.com/GPX/1/1",
            "http://www.topografix.com/GPX/1/1/gpx.xsd",
        ]

    def parse(self) -> Gpx:
        """
        Parse Fit file.

        Returns:
            Gpx: Gpx instance.
        """
        # Parse FIT file
        try:
            self._parse()
        except Exception as err:
            logging.exception(
                "Unexpected %s, %s.\n" "Unable to parse FIT file.", err, type(err)
            )
            raise

        # Add properties
        self._add_properties()

        logging.debug("Parsing complete!!")
        return self.gpx



import errno
import logging
import os
import warnings
import xml.etree.ElementTree as ET
from typing import Dict, List, Optional

from ..gpx_elements import (
    Bounds,
    Copyright,
    Email,
    Extensions,
    Gpx,
    Link,
    Metadata,
    Person,
    Point,
    PointSegment,
    Route,
    Track,
    TrackSegment,
    WayPoint,
)
from .gpx_writer_method_behavior_creator import GPXWriterMethodBehaviorCreator
from .writer import Writer


class GPXWriter(Writer):
    """
    GPX file writer.
    """

    def __init__(
        self, gpx: Gpx = None, precisions: Dict = None, time_format: str = None
    ) -> None:
        """
        Initialize GPXWriter instance.

        Args:
            gpx (Gpx, optional): Gpx instance to write. Defaults to None.
            precisions (dict, optional): Decimal precision for each type of value. Defaults to None.
            time_format (dict, optional): Time format. Defaults to None.
        """
        super().__init__(gpx, precisions, time_format)

        # Utility attributes
        self.file_name: str = ""
        self.gpx_string: str = ""
        self.gpx_root = None

        # Methods behavior creator
        self.behavior_creator: GPXWriterMethodBehaviorCreator = (
            GPXWriterMethodBehaviorCreator()
        )

        # Methods behaviors
        self._add_bounds = self.placeholder_behavior
        self._add_copyright = self.placeholder_behavior
        self._add_email = self.placeholder_behavior
        self._add_link = self.placeholder_behavior
        self._add_metadata = self.placeholder_behavior
        self._add_person = self.placeholder_behavior
        self._add_point_segment = self.placeholder_behavior
        self._add_point = self.placeholder_behavior
        self._add_route = self.placeholder_behavior
        self._add_track_segment = self.placeholder_behavior
        self._add_track = self.placeholder_behavior
        self._add_way_point = self.placeholder_behavior
        self._add_track_point = self.placeholder_behavior

        # Fields
        self.properties: bool = None
        self.bounds_fields: List[str] = None
        self.copyright_fields: List[str] = None
        self.email_fields: List[str] = None
        self.extensions_fields: Dict = None
        self.gpx_fields: List[str] = None
        self.link_fields: List[str] = None
        self.metadata_fields: List[str] = None
        self.person_fields: List[str] = None
        self.point_segment_fields: List[str] = None
        self.point_fields: List[str] = None
        self.route_fields: List[str] = None
        self.track_segment_fields: List[str] = None
        self.track_fields: List[str] = None
        self.way_point_fields: List[str] = None
        self.track_point_fields: List[str] = None

    def placeholder_behavior(self, element, subelement):
        """
        Placeholder function for adding subelement to element in XML
        tree.

        Parameters
        ----------
        element : ET.Element
            Parent element in XML tree
        subelement : GpxElement
            Subelement in XML tree

        Returns
        -------
        ET.Element
            Parent element containing subelement
        """
        print(element, subelement)
        return None

    def add_bounds(self, element: ET.Element, bounds: Bounds) -> ET.Element:
        """
        Add Bounds instance to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            bounds (Bounds): Bounds instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_bounds(self, element, bounds)

    def add_copyright(self, element: ET.Element, copyright_: Copyright) -> ET.Element:
        """
        Add Copyright instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            copyright (Copyright): Copyright instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_copyright(self, element, copyright_)

    def add_email(self, element: ET.Element, email: Email) -> ET.Element:
        """
        Add Email instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            email (Email): Email instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_email(self, element, email)

    def _add_extensions_rec(self, extensions_, values, extensions_fields) -> ET.Element:
        if extensions_ is not None:
            # Add n-th level extensions
            for k0, v0 in values.items():
                if k0 in extensions_fields.keys():
                    # If k0 contains sub-elements
                    if isinstance(v0["elmts"], dict):
                        # Create sub-element
                        sub_extensions_ = ET.SubElement(extensions_, k0)

                        # Set attributes
                        for k1, v1 in v0["attrib"].items():
                            if k1 in extensions_fields[k0]["attrib"].keys():
                                self.setIfNotNone(extensions_, k1, v1)

                        # Add (n+1)-th level extensions
                        sub_extensions_ = self._add_extensions_rec(
                            sub_extensions_, v0["elmts"], extensions_fields[k0]["elmts"]
                        )
                    # Else, k0 contains a value
                    else:
                        extensions_, sub_extensions_ = self.add_subelement(
                            extensions_, k0, v0["elmts"]
                        )

                        # Set attributes
                        for k1, v1 in v0["attrib"].items():
                            if k1 in extensions_fields[k0]["attrib"].keys():
                                self.setIfNotNone(sub_extensions_, k1, v1)

        return extensions_

    def add_extensions(
        self, element: ET.Element, extensions: Extensions, extensions_fields: Dict
    ) -> ET.Element:
        """
        Add Extnsions instance element to GPX element.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            GPX element
        extensions : Extensions
            Extensions instance to add
        extensions_fields : Dict
            Extensions fileds to add

        Returns
        -------
        xml.etree.ElementTree.Element
            GPX element
        """
        if extensions is not None:
            extensions_ = ET.SubElement(element, extensions.tag)
            extensions_ = self._add_extensions_rec(
                extensions_, extensions.values, extensions_fields
            )
        return element

    def add_link(self, element: ET.Element, link: Link) -> ET.Element:
        """
        Add Link instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            link (Link): Link instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_link(self, element, link)

    def add_metadata(self, element: ET.Element, metadata: Metadata) -> ET.Element:
        """
        Add Metadata instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            metadata (Metadata): Metadata instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_metadata(self, element, metadata)

    def add_person(self, element: ET.Element, person: Person) -> ET.Element:
        """
        Add Person instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            person (Person): Person instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_person(self, element, person)

    def add_point_segment(
        self, element: ET.Element, point_segment: PointSegment
    ) -> ET.Element:
        """
        Add PointSegment instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            point_segment (PointSegment): PointSegment instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_point_segment(self, element, point_segment)

    def add_point(self, element: ET.Element, point: Point) -> ET.Element:
        """
        Add Point instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            point (Point): Point instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_point(self, element, point)

    def add_route(self, element: ET.Element, route: Route) -> ET.Element:
        """
        Add Route instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            route (Route): Route instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_route(self, element, route)

    def add_track_segment(
        self, element: ET.Element, track_segment: TrackSegment
    ) -> ET.Element:
        """
        Add TrackSegment instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            track_segment (TrackSegment): TrackSegment instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_track_segment(self, element, track_segment)

    def add_track(self, element: ET.Element, track: Track) -> ET.Element:
        """
        Add Track instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            track (Track): Track instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_track(self, element, track)

    def add_way_point(self, element: ET.Element, way_point: WayPoint) -> ET.Element:
        """
        Add WayPoint instance element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            way_point (WayPoint): WayPoint instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_way_point(self, element, way_point)

    def add_track_point(self, element: ET.Element, way_point: WayPoint) -> ET.Element:
        """
        Add WayPoint instance (with trkpt tag) element to GPX element.

        Args:
            element (xml.etree.ElementTree.Element): GPX element.
            way_point (WayPoint): WayPoint instance to add.

        Returns:
            xml.etree.ElementTree.Element: GPX element.
        """
        return self._add_track_point(self, element, way_point)

    def _create_schema_loc_str(self, xsi_schema_location: list[str]) -> str:
        """
        Create schema location string to write in GPX file.

        Parameters
        ----------
        xsi_schema_location : list[str]
            List of schema locations.

        Returns
        -------
        str
            Schema location string to write in GPX file.
        """
        schema_location_string = ""
        for loc in xsi_schema_location:
            schema_location_string += loc
            schema_location_string += " "
        schema_location_string = schema_location_string[:-1]
        return schema_location_string

    def add_root_properties(self) -> None:
        """
        Add properties to the GPX root element.
        """
        self.setIfNotNone(self.gpx_root, "version", self.gpx.version)
        self.setIfNotNone(self.gpx_root, "creator", "ezGPX")
        for k, v in self.gpx.xmlns.items():
            self.setIfNotNone(self.gpx_root, "xmlns:" + k if k != "" else "xmlns", v)
        schema_location_string = self._create_schema_loc_str(
            self.gpx.xsi_schema_location
        )
        self.setIfNotNone(self.gpx_root, "xsi:schemaLocation", schema_location_string)

    def add_root_metadata(self) -> None:
        """
        Add metadata element to the GPX root element.
        """
        self.gpx_root = self.add_metadata(self.gpx_root, self.gpx.metadata)

    def add_root_way_points(self) -> None:
        """
        Add wpt elements to the GPX root element.
        """
        for way_point in self.gpx.wpt:
            self.gpx_root = self.add_way_point(self.gpx_root, way_point)

    def add_root_routes(self) -> None:
        """
        Add rte elements to the GPX root element.
        """
        for route in self.gpx.rte:
            self.gpx_root = self.add_route(self.gpx_root, route)

    def add_root_tracks(self) -> None:
        """
        Add trck elements to the GPX root element.
        """
        for track in self.gpx.trk:
            self.gpx_root = self.add_track(self.gpx_root, track)

    def add_root_extensions(self) -> None:
        """
        Add extensions element to the GPX root element.
        """
        if self.gpx.extensions is not None:
            self.gpx_root = self.add_extensions(
                self.gpx_root, self.gpx.extensions, self.extensions_fields.get("gpx")
            )

    def gpx_to_string(self) -> str:
        """
        Convert Gpx instance to a string (the content of a .gpx file).

        Returns:
            str: String corresponding to the Gpx instance.
        """
        if self.gpx is not None:
            # Reset string
            self.gpx_string = ""

            # Root
            self.gpx_root = ET.Element("gpx")

            # Properties
            if self.properties:
                self.add_root_properties()

            # Metadata
            if self.metadata_fields:
                self.add_root_metadata()

            # Way points
            if self.way_point_fields:
                self.add_root_way_points()

            # Routes
            if self.route_fields:
                self.add_root_routes()

            # Tracks
            if self.track_fields:
                self.add_root_tracks()

            # Extensions
            if self.extensions_fields.get("gpx"):
                self.add_root_extensions()

            # Convert data to string
            logging.info("Converting GPX to string...")
            self.gpx_string = ET.tostring(self.gpx_root, encoding="unicode")
            # self.gpx_string = ET.tostring(gpx_root, encoding="utf-8")

            logging.info("GPX successfully converted to string:\n%s", self.gpx_string)

            return self.gpx_string

    def write_gpx(self):
        """
        Convert Gpx instance to string and write to file.
        """
        # Open/create GPX file
        try:
            f = open(self.file_path, "w", encoding="utf-8")
        except OSError:
            logging.exception("Could not open/read file: %s", self.file_path)
            raise
        # Write GPX file
        with f:
            f.write('<?xml version="1.0" encoding="UTF-8"?>')
            f.write(self.gpx_string)

    def write(
        self,
        file_path: str,
        properties: bool = True,
        bounds_fields: Optional[List[str]] = None,
        copyright_fields: Optional[List[str]] = None,
        email_fields: Optional[List[str]] = None,
        extensions_fields: Optional[Dict] = None,
        gpx_fields: Optional[List[str]] = None,
        link_fields: Optional[List[str]] = None,
        metadata_fields: Optional[List[str]] = None,
        person_fields: Optional[List[str]] = None,
        point_segment_fields: Optional[List[str]] = None,
        point_fields: Optional[List[str]] = None,
        route_fields: Optional[List[str]] = None,
        track_segment_fields: Optional[List[str]] = None,
        track_fields: Optional[List[str]] = None,
        way_point_fields: Optional[List[str]] = None,
        track_point_fields: Optional[List[str]] = None,
        mandatory_fields: bool = True,
        xml_schema: bool = False,
        xml_extensions_schemas: bool = False,
    ) -> bool:
        """
        TO UPDATE
        Handle writing.

        Args:
            path (str): Path to write the GPX file.
            properties (bool, optional): Toggle properties writting. Defaults to True.
            metadata (bool, optional): Toggle metadata writting. Defaults to True.
            way_point (bool, optional): Toggle way points writting. Defaults to True.
            routes (bool, optional): Toggle routes writting. Defaults to True.
            extensions (bool, optional): Toggle extensions writting. Defaults to True.
            ele (bool, optional): Toggle elevation writting. Defaults to True.
            time (bool, optional): Toggle time writting. Defaults to True.
            check_xml_schemas (bool, optional): Toggle schema verification after writting. Defaults to False.
            extensions_schemas (bool, optional): Toggle extensions schema verificaton after writing. Requires internet connection and is not guaranted to work. Defaults to False.

        Returns:
            bool: Return False if written file does not follow checked schemas. Return True otherwise.
        """
        directory_path = os.path.dirname(os.path.realpath(file_path))
        if not os.path.exists(directory_path):
            raise NotADirectoryError(
                errno.ENOENT, os.strerror(errno.ENOENT), directory_path
            )
        self.file_path = file_path
        self.file_name = os.path.basename(self.file_path)

        # Set parameters
        self.properties = properties
        self.bounds_fields = (
            bounds_fields if bounds_fields is not None else Bounds.fields
        )
        self.copyright_fields = (
            copyright_fields if copyright_fields is not None else Copyright.fields
        )
        self.email_fields = email_fields if email_fields is not None else Email.fields
        self.extensions_fields = (
            extensions_fields if extensions_fields is not None else {}
        )
        self.gpx_fields = gpx_fields if gpx_fields is not None else Gpx.fields
        self.link_fields = link_fields if link_fields is not None else Link.fields
        self.metadata_fields = (
            metadata_fields if metadata_fields is not None else Metadata.fields
        )
        self.person_fields = (
            person_fields if person_fields is not None else Person.fields
        )
        self.point_segment_fields = (
            point_segment_fields
            if point_segment_fields is not None
            else PointSegment.fields
        )
        self.point_fields = point_fields if point_fields is not None else Point.fields
        self.route_fields = route_fields if route_fields is not None else Route.fields
        self.track_segment_fields = (
            track_segment_fields
            if track_segment_fields is not None
            else TrackSegment.fields
        )
        self.track_fields = track_fields if track_fields is not None else Track.fields
        self.way_point_fields = (
            way_point_fields if way_point_fields is not None else WayPoint.fields
        )
        self.track_point_fields = (
            track_point_fields if track_point_fields is not None else WayPoint.fields
        )

        # Check mandatory fields
        if mandatory_fields:

            def check_mandatory_fields(element, fields, mandatory_fields):
                if any(f not in fields for f in mandatory_fields):
                    warnings.warn(
                        f"{element} element must have following fields: {mandatory_fields}"
                        "Missing mandatory fields will automatically be added."
                    )
                    fields = list(set(fields + mandatory_fields))
                return fields

            self.bounds_fields = check_mandatory_fields(
                "Bounds", self.bounds_fields, Bounds.mandatory_fields
            )
            self.copyright_fields = check_mandatory_fields(
                "Copyright", self.copyright_fields, Copyright.mandatory_fields
            )
            self.email_fields = check_mandatory_fields(
                "Email", self.email_fields, Email.mandatory_fields
            )
            self.gpx_fields = check_mandatory_fields(
                "Gpx", self.gpx_fields, Gpx.mandatory_fields
            )
            self.link_fields = check_mandatory_fields(
                "Link", self.link_fields, Link.mandatory_fields
            )
            self.metadata_fields = check_mandatory_fields(
                "Metadata", self.metadata_fields, Metadata.mandatory_fields
            )
            self.person_fields = check_mandatory_fields(
                "Person", self.person_fields, Person.mandatory_fields
            )
            self.point_segment_fields = check_mandatory_fields(
                "PointSegment", self.point_segment_fields, PointSegment.mandatory_fields
            )
            self.point_fields = check_mandatory_fields(
                "Point", self.point_fields, Point.mandatory_fields
            )
            self.route_fields = check_mandatory_fields(
                "Route", self.route_fields, Route.mandatory_fields
            )
            self.track_segment_fields = check_mandatory_fields(
                "TrackSegment", self.track_segment_fields, TrackSegment.mandatory_fields
            )
            self.track_fields = check_mandatory_fields(
                "Track", self.track_fields, Track.mandatory_fields
            )
            self.way_point_fields = check_mandatory_fields(
                "WayPoint", self.way_point_fields, WayPoint.mandatory_fields
            )
            self.track_point_fields = check_mandatory_fields(
                "TrackPoint", self.track_point_fields, WayPoint.mandatory_fields
            )

        # Create methods behaviors
        self._add_bounds = self.behavior_creator.add_bounds_creator(self.bounds_fields)
        self._add_copyright = self.behavior_creator.add_copyright_creator(
            self.copyright_fields
        )
        self._add_email = self.behavior_creator.add_email_creator(self.email_fields)
        self._add_link = self.behavior_creator.add_link_creator(self.link_fields)
        self._add_metadata = self.behavior_creator.add_metadata_creator(
            self.metadata_fields
        )
        self._add_person = self.behavior_creator.add_person_creator(self.person_fields)
        self._add_point_segment = self.behavior_creator.add_point_segment_creator(
            self.point_segment_fields
        )
        self._add_point = self.behavior_creator.add_point_creator(self.point_fields)
        self._add_route = self.behavior_creator.add_route_creator(self.route_fields)
        self._add_track_segment = self.behavior_creator.add_track_segment_creator(
            self.track_segment_fields
        )
        self._add_track = self.behavior_creator.add_track_creator(self.track_fields)
        self._add_way_point = self.behavior_creator.add_way_point_creator(
            self.way_point_fields
        )
        self._add_track_point = self.behavior_creator.add_track_point_creator(
            self.track_point_fields
        )

        # Write .gpx file
        self.gpx_to_string()
        self.write_gpx()

        # Check XML schemas
        res = True
        if xml_schema or xml_extensions_schemas:
            res = self.check_xml_schemas(xml_schema, xml_extensions_schemas)

        return res



from .gpx_element import GpxElement


class Copyright(GpxElement):
    """
    copyrightType element in GPX file.
    """

    fields = ["author", "year", "licence"]
    mandatory_fields = ["author"]

    def __init__(
        self,
        tag: str = "copyright",
        author: str = None,
        year: int = None,
        licence: str = None,
    ) -> None:
        """
        Initialize Copyright instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "copyright"
        author : str, optional
            Author, by default None
        year : int, optional
            Year, by default None
        licence : str, optional
            Licence, by default None
        """
        self.tag: str = tag
        self.author: str = author
        self.year: int = year
        self.licence: str = licence



try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

import logging
from datetime import datetime, timezone
from typing import Dict, List, Tuple, Type, Union

import pandas as pd
import polars as pl
import xmlschema

from ..utils import haversine_distance, ramer_douglas_peucker
from .extensions import Extensions
from .gpx_element import GpxElement
from .metadata import Metadata
from .route import Route
from .track import Track
from .way_point import WayPoint


class Gpx(GpxElement):
    """
    gpxType element in GPX file.
    """

    fields = ["version", "creator", "metadata", "wpt", "rte", "trk", "extensions"]
    mandatory_fields = ["version", "creator"]

    def __init__(
        self,
        tag: str = "gpx",
        version: str = None,
        creator: str = None,
        xsi_schema_location: List[str] = None,
        xmlns: Dict = None,
        metadata: Metadata = None,
        wpt: List[WayPoint] = None,
        rte: List[Route] = None,
        trk: List[Track] = None,
        extensions: Extensions = None,
    ) -> None:

        self.tag: str = tag
        self.version: str = version
        self.creator: str = creator
        if xsi_schema_location is None:
            self.xsi_schema_location: List[str] = []
        else:
            self.xsi_schema_location: List[str] = xsi_schema_location
        if xmlns is None:
            self.xmlns: Dict = {}
        else:
            self.xmlns: Dict = xmlns
        self.metadata: Metadata = metadata
        if wpt is None:
            self.wpt: List[WayPoint] = []
        else:
            self.wpt: List[WayPoint] = wpt
        if rte is None:
            self.rte: List[Route] = []
        else:
            self.rte: List[Route] = rte
        if trk is None:
            self.trk: List[Track] = []
        else:
            self.trk: List[Track] = trk
        self.extensions: Extensions = extensions

    ###############################################################################
    #### Schemas ##################################################################
    ###############################################################################

    def check_xml_schema(self, file_path: str) -> bool:
        """
        Check XML schema.

        Parameters
        ----------
        file_path : str
            File path

        Returns
        -------
        bool
            True if the file follows XML schemas
        """
        schema = None

        # GPX
        if file_path.endswith(".gpx"):
            if self.version == "1.1":
                schema = xmlschema.XMLSchema(
                    files("ezgpx.schemas").joinpath("gpx_1_1/gpx.xsd")
                )
            elif self.version == "1.0":
                schema = xmlschema.XMLSchema(
                    files("ezgpx.schemas").joinpath("gpx_1_0/gpx.xsd")
                )
            else:
                logging.error("Unable to check XML schema (unsupported GPX version)")
                return False

        # KML
        elif file_path.endswith(".kml"):
            schema = xmlschema.XMLSchema(
                files("ezgpx.schemas").joinpath("kml_2_2/ogckml22.xsd")
            )

        # KMZ
        elif file_path.endswith(".kmz"):
            return False

        # FIT
        elif file_path.endswith(".fit"):
            logging.error("Unable to check XML schema (fit files are not XML files)")
            return False

        # NOT SUPPORTED
        else:
            logging.error("Unable to check XML schema (unable to identify file type)")
            return False

        if schema is not None:
            return schema.is_valid(file_path)
        else:
            logging.error("Unable to check XML schema (unable to load XML schema)")
            return False

    def check_xml_extensions_schemas(self, file_path: str) -> bool:
        """
        Check XML extensions schemas.

        Parameters
        ----------
        file_path : str
            Path to GPX file.

        Returns
        -------
        bool
            True if GPX file follows XML extensions schemas.
        """
        gpx_schemas = [s for s in self.xsi_schema_location if s.endswith(".xsd")]
        gpx_schemas.remove("http://www.topografix.com/GPX/1/1/gpx.xsd")
        for gpx_schema in gpx_schemas:
            logging.debug("schema = %s", gpx_schema)
            schema = xmlschema.XMLSchema(gpx_schema)
            if not schema.is_valid(file_path):
                logging.error("File does not follow %s", gpx_schema)
                return False

    ###############################################################################
    #### Name #####################################################################
    ###############################################################################

    def name(self) -> str:
        """
        Return activity name.

        Returns
        -------
        str
            Activity name.
        """
        return self.trk[0].name

    def set_name(self, new_name: str) -> None:
        """
        Set name.

        Parameters
        ----------
        new_name : str
            New name.
        """
        self.trk[0].name = new_name

    ###############################################################################
    #### Points ###################################################################
    ###############################################################################

    def nb_points(self) -> int:
        """
        Compute the number of points in the GPX.

        Returns
        -------
        int
            Number of points in the GPX.
        """
        nb_pts = 0
        for track in self.trk:
            for track_segment in track.trkseg:
                nb_pts += len(track_segment.trkpt)
        return nb_pts

    def bounds(self) -> Tuple[float, float, float, float]:
        """
        Find minimum and maximum latitude and longitude.

        Returns
        -------
        Tuple[float, float, float, float]
            Min latitude, min longitude, max latitude, max longitude.
        """
        min_lat = self.trk[0].trkseg[0].trkpt[0].lat
        min_lon = self.trk[0].trkseg[0].trkpt[0].lon
        max_lat = min_lat
        max_lon = min_lon

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.lat < min_lat:
                        min_lat = track_point.lat
                    if track_point.lon < min_lon:
                        min_lon = track_point.lon
                    if track_point.lat > max_lat:
                        max_lat = track_point.lat
                    if track_point.lon > max_lon:
                        max_lon = track_point.lon
        return min_lat, min_lon, max_lat, max_lon

    def center(self) -> Tuple[float, float]:
        """
        Compute the center coordinates of the track.

        Returns
        -------
        Tuple[float, float]
            Latitude and longitude of the center point.
        """
        min_lat, min_lon, max_lat, max_lon = self.bounds()
        center_lat = min_lat + (max_lat - min_lat) / 2
        center_lon = min_lon + (max_lon - min_lon) / 2
        return center_lat, center_lon

    def first_point(self) -> WayPoint:
        """
        Return GPX first point.

        Returns
        -------
        WayPoint
            First point.
        """
        return self.trk[0].trkseg[0].trkpt[0]

    def last_point(self) -> WayPoint:
        """
        Return GPX last point.

        Returns
        -------
        WayPoint
            Last point.
        """
        return self.trk[-1].trkseg[-1].trkpt[-1]

    def extreme_points(self) -> Tuple[WayPoint, WayPoint, WayPoint, WayPoint]:
        """
        Find extreme points in track, i.e.: points with lowest and highest latitude and longitude.

        Returns
        -------
        Tuple[WayPoint, WayPoint, WayPoint, WayPoint]
            Min latitude point, min longitude point, max latitude point, max longitude point
        """
        min_lat_point = self.trk[0].trkseg[0].trkpt[0]
        min_lon_point = self.trk[0].trkseg[0].trkpt[0]
        max_lat_point = self.trk[0].trkseg[0].trkpt[0]
        max_lon_point = self.trk[0].trkseg[0].trkpt[0]

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.lat < min_lat_point.lat:
                        min_lat_point = track_point
                    if track_point.lon < min_lon_point.lon:
                        min_lon_point = track_point
                    if track_point.lat > max_lat_point.lat:
                        max_lat_point = track_point
                    if track_point.lon > max_lon_point.lon:
                        max_lon_point = track_point
        return min_lat_point, min_lon_point, max_lat_point, max_lon_point

    ###############################################################################
    #### Distance and Elevation ###################################################
    ###############################################################################

    def distance(self) -> float:
        """
        Compute the total distance (meters) of tracks contained in the Gpx element.

        Returns
        -------
        float
            Distance (meters).
        """
        dst = 0.0
        previous_point = self.trk[0].trkseg[0].trkpt[0]
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    dst += haversine_distance(previous_point, track_point)
                    previous_point = track_point
        return dst

    def compute_points_distance_from_start(self):
        """
        Compute distance from start at each point.
        """
        dst = 0.0
        previous_point = self.trk[0].trkseg[0].trkpt[0]
        previous_point.distance_from_start = dst
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    dst += haversine_distance(previous_point, track_point)
                    track_point.distance_from_start = dst
                    previous_point = track_point

    def ascent(self) -> float:
        """
        Compute the total ascent (meters) of tracks contained in the Gpx element.

        Returns
        -------
        float
            Ascent (meters).
        """
        ascent = 0
        previous_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele > previous_elevation:
                        ascent += track_point.ele - previous_elevation
                    previous_elevation = track_point.ele
        return ascent

    def descent(self) -> float:
        """
        Compute the total descent (meters) of tracks contained in the Gpx element.

        Returns
        -------
        float
            Descent (meters).
        """
        descent = 0
        previous_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele < previous_elevation:
                        descent += previous_elevation - track_point.ele
                    previous_elevation = track_point.ele
        return descent

    def min_elevation(self) -> float:
        """
        Compute minimum elevation (meters) in tracks contained in the Gpx element.

        Returns
        -------
        float
            Minimum elevation (meters).
        """
        min_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele < min_elevation:
                        min_elevation = track_point.ele
        return min_elevation

    def max_elevation(self) -> float:
        """
        Compute maximum elevation (meters) in tracks contained in the Gpx element.

        Returns
        -------
        float
            Maximum elevation (meters).
        """
        max_elevation = self.trk[0].trkseg[0].trkpt[0].ele
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ele > max_elevation:
                        max_elevation = track_point.ele
        return max_elevation

    def compute_points_ascent_rate(self) -> None:
        """
        Compute ascent rate at each point.
        """
        previous_point = self.first_point()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    distance = haversine_distance(previous_point, track_point)
                    ascent = track_point.ele - previous_point.ele
                    try:
                        track_point.ascent_rate = (ascent * 100) / distance
                    except ZeroDivisionError:
                        track_point.ascent_rate = 0.0
                    previous_point = track_point

    def min_ascent_rate(self) -> float:
        """
        Return activity minimum ascent rate.

        Returns
        -------
        float
            Minimum ascent rate.
        """
        min_ascent_rate = 100.0
        self.compute_points_ascent_rate()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_rate < min_ascent_rate:
                        min_ascent_rate = track_point.ascent_rate

        return min_ascent_rate

    def max_ascent_rate(self) -> float:
        """
        Return activity maximum ascent rate.

        Returns
        -------
        float
            Maximum ascent rate.
        """
        max_ascent_rate = -1.0
        self.compute_points_ascent_rate()  # Check if it needs to be done

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_rate > max_ascent_rate:
                        max_ascent_rate = track_point.ascent_rate

        return max_ascent_rate

    ###############################################################################
    #### Time #####################################################################
    ###############################################################################

    def utc_start_time(self) -> datetime:
        """
        Return the activity UTC start time.

        Returns
        -------
        datetime
            UTC start time.
        """
        return self.trk[0].trkseg[0].trkpt[0].time

    def utc_stop_time(self):
        """
        Return the activity UTC stop time.

        Returns
        -------
        datetime
            UTC stop time.
        """
        return self.trk[-1].trkseg[-1].trkpt[-1].time

    def start_time(self) -> datetime:
        """
        Return the activity start time.

        Returns
        -------
        datetime
            Start time.
        """
        start_time = None
        try:
            start_time = (
                self.trk[0]
                .trkseg[0]
                .trkpt[0]
                .time.replace(tzinfo=timezone.utc)
                .astimezone(tz=None)
            )
        except:
            logging.error("Unable to find activity start time")
        return start_time

    def stop_time(self) -> datetime:
        """
        Return the activity stop time.

        Returns
        -------
        datetime
            Stop time.
        """
        stop_time = None
        try:
            stop_time = (
                self.trk[-1]
                .trkseg[-1]
                .trkpt[-1]
                .time.replace(tzinfo=timezone.utc)
                .astimezone(tz=None)
            )
        except:
            logging.error("Unable to find activity stop time")
        return stop_time

    def total_elapsed_time(self) -> datetime:
        """
        Compute the total elapsed time.

        Returns
        -------
        datetime
            Total elapsed time.
        """
        total_elapsed_time = None
        try:
            total_elapsed_time = self.stop_time() - self.start_time()
        except:
            logging.error("Unable to compute activity total elapsed time")
        return total_elapsed_time

    def stopped_time(self, tolerance: float = 2.45) -> datetime:
        """
        Compute the stopped time during activity.

        Parameters
        ----------
        tolerance : float, optional
            Maximal distance between two points for movement, by default 2.45
            (According to my tests with strava_run_1.gpx and the data on Strava)

        Returns
        -------
        datetime
            Stopped time.
        """
        stopped_time = self.start_time() - self.start_time()  # Better way to do it?

        previous_point = self.trk[0].trkseg[0].trkpt[0]

        for track in self.trk:
            for segment in track.trkseg:
                for point in segment.trkpt:
                    if haversine_distance(previous_point, point) < tolerance:
                        stopped_time += point.time - previous_point.time
                    previous_point = point

        return stopped_time

    def moving_time(self) -> datetime:
        """
        Compute the moving time during the activity.

        Returns
        -------
        datetime
            Moving time.
        """
        return self.total_elapsed_time() - self.stopped_time()

    ###############################################################################
    #### Speed and Pace ###########################################################
    ###############################################################################

    def avg_speed(self) -> float:
        """
        Compute the average speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Average speed (kilometres per hour).
        """
        # Compute and convert total elapsed time
        total_elapsed_time = self.total_elapsed_time()
        total_elapsed_time = total_elapsed_time.total_seconds() / 3600

        # Compute and convert distance
        distance = self.distance() / 1000

        return distance / total_elapsed_time

    def avg_moving_speed(self) -> float:
        """
        Compute the average moving speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Average moving speed (kilometres per hour).
        """
        # Compute and convert moving time
        moving_time = self.moving_time()
        moving_time = moving_time.total_seconds() / 3600

        # Compute and convert distance
        distance = self.distance() / 1000

        return distance / moving_time

    def compute_points_speed(self) -> None:
        """
        Compute speed (kilometres per hour) at each track point.
        """
        previous_point = self.first_point()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    distance = (
                        haversine_distance(previous_point, track_point) / 1000
                    )  # Convert to kilometers
                    # Convert to hours
                    time = (
                        track_point.time - previous_point.time
                    ).total_seconds() / 3600
                    try:
                        track_point.speed = distance / time
                    except ZeroDivisionError:
                        track_point.speed = 0.0
                    previous_point = track_point

    def min_speed(self) -> float:
        """
        Return the minimum speed during the activity.

        Returns
        -------
        float
            Minimum speed.
        """
        min_speed = 1000.0
        self.compute_points_speed()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.speed < min_speed:
                        min_speed = track_point.speed

        return min_speed

    def max_speed(self) -> float:
        """
        Return the maximum speed during the activity.

        Returns
        -------
        float
            Maximum speed.
        """
        max_speed = -1.0
        self.compute_points_speed()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.speed > max_speed:
                        max_speed = track_point.speed

        return max_speed

    def avg_pace(self) -> float:
        """
        Compute the average pace (minute per kilometer) during the activity.

        Returns
        -------
        float
            Average pace (minute per kilometer).
        """
        return 60.0 / self.avg_speed()

    def avg_moving_pace(self) -> float:
        """
        Compute the average moving pace (minute per kilometer) during the activity.

        Returns
        -------
        float
            Average moving pace (minute per kilometer).
        """
        return 60.0 / self.avg_moving_speed()

    def compute_points_pace(self) -> None:
        """
        Compute pace at each track point.
        """
        self.compute_points_speed()

        for track in self.trk:
            for segment in track.trkseg:
                for point in segment.trkpt:
                    try:
                        point.pace = 60.0 / point.speed
                    except ZeroDivisionError:
                        # Fill with average moving pace (first point)
                        point.pace = self.avg_moving_pace()

    def min_pace(self) -> float:
        """
        Return the minimum pace during the activity.

        Returns
        -------
        float
            Minimum pace.
        """
        min_pace = 1000.0
        self.compute_points_pace()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.pace < min_pace:
                        min_pace = track_point.pace

        return min_pace

    def max_pace(self) -> float:
        """
        Return the maximum pace during the activity.

        Returns
        -------
        float
            Maximum pace.
        """
        max_pace = -1.0
        self.compute_points_pace()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.pace > max_pace:
                        max_pace = track_point.pace

        return max_pace

    def compute_points_ascent_speed(self) -> None:
        """
        Compute ascent speed (kilometres per hour) at each track point.
        """
        previous_point = self.first_point()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    ascent = (
                        track_point.ele - previous_point.ele
                    ) / 1000  # Convert to kilometers
                    # Convert to hours
                    time = (
                        track_point.time - previous_point.time
                    ).total_seconds() / 3600
                    try:
                        track_point.ascent_speed = ascent / time
                    except ZeroDivisionError:
                        track_point.ascent_speed = 0.0
                    previous_point = track_point

    def min_ascent_speed(self) -> float:
        """
        Return the minimum ascent speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Minimum ascent speed.
        """
        min_ascent_speed = 1000.0
        self.compute_points_ascent_speed()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_speed < min_ascent_speed:
                        min_ascent_speed = track_point.ascent_speed

        return min_ascent_speed

    def max_ascent_speed(self) -> float:
        """
        Return the maximum ascent speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Maximum ascent speed.
        """
        max_ascent_speed = -1.0
        self.compute_points_ascent_speed()

        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if track_point.ascent_speed > max_ascent_speed:
                        max_ascent_speed = track_point.ascent_speed

        return max_ascent_speed

    ###############################################################################
    #### Data Removal #############################################################
    ###############################################################################

    def remove_metadata(self):
        """
        Remove metadata.
        """
        self.metadata = None

    def remove_elevation(self):
        """
        Remove elevation data.
        """
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    track_point.ele = None

    def remove_time(self):
        """
        Remove time data.
        """
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    track_point.time = None

    def remove_extensions(self):
        """
        Remove extensions data.
        """
        # Remove extensions from gpx
        self.extensions = None

        # Remove extensions from metadata
        if self.metadata is not None:
            self.metadata.extensions = None

        # Remove extensions from waypoints
        if self.wpt is not None:
            for pt in self.wpt:
                pt.extensions = None

        # Remove extensions from routes
        if self.rte is not None:
            for rt in self.rte:
                rt.extensions = None

        # Remove extensions from tracks
        if self.trk is not None:
            for track in self.trk:
                track.extensions = None
                # Remove extensions from track segments
                if track.trkseg is not None:
                    for track_segment in track.trkseg:
                        track_segment.extensions = None
                        if track_segment.trkpt is not None:
                            for track_point in track_segment.trkpt:
                                # Remove extensions from track points
                                track_point.extensions = None

    ###############################################################################
    #### Error Correction #########################################################
    ###############################################################################

    def remove_points(self, remove_factor: int = 2):
        count = 0
        for track in self.trk:
            for track_segment in track.trkseg:
                for track_point in track_segment.trkpt:
                    if count % remove_factor == 0:
                        track_segment.trkpt.remove(track_point)
                        count += 1

    def remove_gps_errors(self, error_distance=100):
        """
        Remove GPS errors.

        Parameters
        ----------
        error_distance : int, optional
            Error threshold distance (meters) between two points, by default 100

        Returns
        -------
        _type_
            List of removed points (GPS errors).
        """
        previous_point = None
        gps_errors = []

        for track in self.trk:
            for track_segment in track.trkseg:

                new_trkpt = []

                for track_point in track_segment.trkpt:
                    # GPS error
                    dst = haversine_distance(previous_point, track_point)
                    if previous_point is not None and dst > error_distance:
                        logging.warning(
                            "Point %s has been removed (GPS error)", track_point
                        )
                        gps_errors.append(track_point)
                    # No GPS error
                    else:
                        new_trkpt.append(track_point)
                        previous_point = track_point

                track_segment.trkpt = new_trkpt

        return gps_errors

    def remove_close_points(self, min_dist: float = 1, max_dist: float = 10):
        """
        Remove points that are to close together.

        Parameters
        ----------
        min_dist : float, optional
            Minimal distance between two points, by default 1
        max_dist : float, optional
            Maximal distance between two points, by default 10
        """
        point_1 = None
        point_2 = None

        for track in self.trk:
            for segment in track.trkseg:

                new_trkpt = []

                for point in segment.trkpt:
                    if point_1 is None:
                        point_1 = point
                        new_trkpt.append(point_1)
                    elif point_2 is None:
                        point_2 = point
                    else:
                        if (
                            haversine_distance(point_1, point_2) < min_dist
                            or haversine_distance(point_2, point) < min_dist
                        ) and haversine_distance(point_1, point) < max_dist:
                            point_2 = point
                        else:
                            new_trkpt.append(point_2)
                            point_1 = point_2
                            point_2 = point

                segment.trkpt = new_trkpt

    ###############################################################################
    #### Simplification ###########################################################
    ###############################################################################

    def simplify(self, epsilon):
        """
        Simplify GPX trk using Ramer-Douglas-Peucker algorithm.

        Parameters
        ----------
        epsilon : _type_
            Tolerance.
        """
        for track in self.trk:
            for segment in track.trkseg:
                segment.trkpt = ramer_douglas_peucker(segment.trkpt, epsilon)

    ###############################################################################
    #### Exports ##################################################################
    ###############################################################################

    def to_dict(
        self,
        values: List[str] = None,
        orient: str = "dict",
        into: Type[dict] = dict,
        index: bool = True,
    ) -> Dict:
        """
        Convert GPX object to dictionary.
        Pandas.DataFrame.to_dict documentation: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html

        Parameters
        ----------
        values : List[str], optional
            values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"
        orient : str, optional
            Same as in Pandas.DataFrame.to_dict, by default "dict"
        into : Type[dict], optional
            Same as in Pandas.DataFrame.to_dict, by default dict
        index : bool, optional
            Same as in Pandas.DataFrame.to_dict, by default True

        Returns
        -------
        Dict
            Return a dictionary representing the GPX. The resulting
            transformation depends on the `orient` parameter.
        """
        df = self.to_pandas(values)
        return df.to_dict(orient, into, index)

    def _to_dict_df(self, values: List[str] = None) -> Dict:
        """
        Convert GPX object to dictionary.

        Parameters
        ----------
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"

        Returns
        -------
        Dict
            Dictionary containing data from GPX.
        """
        # Set default parameter
        if values is None:
            values = ["lat", "lon"]

        # Compute required values
        test_point = self.first_point()
        if "speed" in values and test_point.speed is None:
            self.compute_points_speed()
        if "pace" in values and test_point.pace is None:
            self.compute_points_pace()
        if "ascent_rate" in values and test_point.ascent_rate is None:
            self.compute_points_ascent_rate()
        if "ascent_speed" in values and test_point.ascent_speed is None:
            self.compute_points_ascent_speed()
        if "distance_from_start" in values and test_point.distance_from_start is None:
            self.compute_points_distance_from_start()

        # Create dataframe
        gpx_data = {}
        for v in values:
            if v == "time":
                gpx_data[v] = [
                    str(trkpt.time.replace(tzinfo=timezone.utc).astimezone(tz=None))
                    for trk in self.trk
                    for trkseg in trk.trkseg
                    for trkpt in trkseg.trkpt
                ]
            else:
                gpx_data[v] = [
                    getattr(trkpt, v)
                    for trk in self.trk
                    for trkseg in trk.trkseg
                    for trkpt in trkseg.trkpt
                ]
        return gpx_data

    def to_pandas(self, values: List[str] = None) -> pd.DataFrame:
        """
        Convert GPX object to Pandas Dataframe.
        Missing values are filled with default values (0 for numerical
        values and empty string for text).

        Parameters
        ----------
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"

        Returns
        -------
        pd.DataFrame
            Dataframe containing data from GPX.
        """
        return pd.DataFrame(self._to_dict_df(values))

    def to_polars(self, values: List[str] = None) -> pl.DataFrame:
        """
        Convert GPX object to Polars Dataframe.
        Missing values are filled with default values (0 for numerical
        values and empty string for text).

        Parameters
        ----------
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"

        Returns
        -------
        pl.DataFrame
            Dataframe containing data from GPX.
        """
        return pl.DataFrame(self._to_dict_df(values))

    def to_csv(
        self,
        path: str = None,
        values: List[str] = None,
        sep: str = ",",
        header: bool = True,
        index: bool = False,
    ) -> Union[str, None]:
        """
        Write the GPX object track coordinates to a .csv file.

        Parameters
        ----------
        path : str, optional
            Path to the .csv file, by default None
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"
        sep : str, optional
            Separator, by default ","
        header : bool, optional
            Toggle header, by default True
        index : bool, optional
             Toggle index, by default False

        Returns
        -------
        str
            CSV like string if path is set to None.
        """
        if values is None:
            values = ["lat", "lon"]

        # Argument columns is required for KML writer (keep values order)
        return self.to_pandas(values).to_csv(
            path, sep=sep, columns=values, header=header, index=index
        )



import io
import os
from typing import Optional

import numpy as np
import pandas as pd
import PIL
import plotly.express as px

from .plotter import Plotter


class PlotlyAnimPlotter(Plotter):
    """
    GPX animated plotter based on Plotly.
    """

    def plot(
        self,
        tiles: str = "open-street-map",  # "open-street-map"
        color: str = "#FFA800",
        title: Optional[str] = None,
        zoom: float = 12.0,
        file_path: Optional[str] = None,
    ):
        """
        Plot (animation) GPX using Plotly.

        Args:
            tiles (str, optional): Map tiles to use. Defaults to
                "open-street-map".
            title (Optional[str], optional): Title of the plot.
                Defaults to None.
            zoom (float, optional): Zoom. Defaults to 12.0.
            file_path (Optional[str], optional): Path to save the plot. Defaults to None.

        Raises:
            FileNotFoundError: Provided path does not exist.

        Returns:
            plotly.Figure: Animated plot of the GPX.
        """
        self._dataframe = self._gpx.to_pandas(["lat", "lon", "ele"])
        center_lat, center_lon = self._gpx.center()
        start = 0
        n_points = len(self._dataframe)

        # Create dataframe for animation
        df = pd.DataFrame()
        for i in np.arange(start, n_points):
            dfa = self._dataframe.head(i).copy()
            dfa["frame"] = i
            df = pd.concat([df, dfa])

        # plotly figure
        fig = px.line_map(
            df, lat="lat", lon="lon", color=color, animation_frame="frame"
        )

        # attribute adjusments
        fig.layout.updatemenus[0].buttons[0]["args"][1]["frame"]["redraw"] = True
        fig.layout.updatemenus[0].buttons[0]["args"][1]["frame"]["duration"] = 0.25
        fig.layout.updatemenus[0].buttons[0]["args"][1]["transition"]["duration"] = 0.25

        # Update layout for nice display
        fig.update_layout(
            margin={"l": 0, "t": 0, "b": 0, "r": 0},
            map={
                "center": {"lon": center_lon, "lat": center_lat},
                "style": tiles,
                "zoom": zoom,
            },
            title={"text": title},
        )

        # Save plot
        if file_path is not None:
            # Check if provided path exists
            directory_path = os.path.dirname(os.path.realpath(file_path))
            if not os.path.exists(directory_path):
                raise FileNotFoundError("Provided path does not exist")
            if file_path.endswith(".html"):
                fig.write_html(file_path)
            elif file_path.endswith(".gif"):
                # Generate image for each step in animation
                frames = []
                for s, fr in enumerate(fig.frames):
                    # Set main traces to appropriate traces within plotly frame
                    fig.update(data=fr.data)
                    # Move slider to correct place
                    fig.layout.sliders[0].update(active=s)
                    # Generate image of current state
                    frames.append(
                        PIL.Image.open(io.BytesIO(fig.to_image(format="png")))
                    )
                # Create animated GIF
                frames[0].save(
                    "test.gif",
                    save_all=True,
                    append_images=frames[1:],
                    optimize=True,
                    duration=500,
                    loop=0,
                )

        return fig



import math
import os
import sys

file_folder = os.path.dirname(__file__)
parent_folder = os.path.realpath(os.path.dirname(file_folder))  # ie: ezGPX

os.chdir(file_folder)
sys.path.append(parent_folder + "/ezgpx")

from ezgpx import WayPoint, utils


class TestUtils:
    def test_haversine_distance(self):
        point_1 = WayPoint("wpt", 48.0, 2.0)
        point_2 = WayPoint("wpt", 43.0, 5.0)
        assert math.isclose(
            utils.haversine_distance(point_1, point_2), 603020.0, abs_tol=1000.0
        )

    def _test_perpendicular_distance_horizontal_line(self):
        start = WayPoint("wpt", 0, 0)
        end = WayPoint("wpt", 0, 2)
        point = WayPoint("wpt", 1, 1)
        return math.isclose(utils.perpendicular_distance(start, end, point), 1)

    def _test_perpendicular_distance_vertical_line(self):
        start = WayPoint("wpt", 0, 0)
        end = WayPoint("wpt", 2, 0)
        point = WayPoint("wpt", 1, 1)
        return math.isclose(utils.perpendicular_distance(start, end, point), 1)

    def _test_perpendicular_distance_random_line(self):
        start = WayPoint("wpt", 0, 0)
        end = WayPoint("wpt", 1, 1)
        point = WayPoint("wpt", 1, 0)
        return math.isclose(
            utils.perpendicular_distance(start, end, point), math.sqrt(2) / 2
        )

    def _test_perpendicular_distance_point_on_line(self):
        start = WayPoint("wpt", 0, 0)
        end = WayPoint("wpt", 1, 1)
        point = WayPoint("wpt", 2, 2)
        return math.isclose(utils.perpendicular_distance(start, end, point), 0)

    def test_perpendicular_distance(self):
        assert self._test_perpendicular_distance_horizontal_line()
        assert self._test_perpendicular_distance_vertical_line()
        assert self._test_perpendicular_distance_random_line()
        assert self._test_perpendicular_distance_point_on_line()



import logging
import xml.etree.ElementTree as ET
from datetime import datetime, timezone
from typing import Dict, Tuple, Union

from ..gpx_elements import Gpx
from ..parsers import DEFAULT_PRECISION, DEFAULT_PRECISION_DICT, DEFAULT_TIME_FORMAT


class Writer:
    """
    File writer.
    """

    def __init__(
        self, gpx: Gpx = None, precisions: Dict = None, time_format: str = None
    ) -> None:
        """
        Initialize Writer instance.

        Parameters
        ----------
        gpx : Gpx, optional
            Gpx instance to write, by default None
        """
        self.gpx: Gpx = gpx

        self.precisions: Dict = (
            precisions if precisions is not None else DEFAULT_PRECISION_DICT
        )
        self.time_format: str = (
            time_format if time_format is not None else DEFAULT_TIME_FORMAT
        )

        self.file_path: str = None

    def setIfNotNone(self, element: ET.Element, field: str, value):
        """
        _summary_

        Parameters
        ----------
        element : ET.Element
            _description_
        field : str
            _description_
        value : _type_
            _description_
        """
        if value is not None:
            element.set(field, value)

    def add_subelement(
        self, element: ET.Element, sub_element: str, text: str
    ) -> Tuple[ET.Element, Union[ET.Element, None]]:
        """
        Add sub-element to GPX element.

        Args:
            element (ET.Element): GPX element.
            sub_element (str): GPX sub-element.
            text (str): GPX sub-element text.

        Returns:
            Tuple[ET.Element, Union[ET.Element, None]]: GPX element and
                GPX sub-element (if not None).
        """
        sub_element_ = None
        if text is not None:
            sub_element_ = ET.SubElement(element, sub_element)
            sub_element_.text = text
        return element, sub_element_

    def add_subelement_number(
        self,
        element: ET.Element,
        sub_element: str,
        number: Union[int, float],
        precision: int = DEFAULT_PRECISION,
    ) -> Tuple[ET.Element, Union[ET.Element, None]]:
        """
        Add sub-element to GPX element.

        Args:
            element (ET.Element): GPX element.
            sub_element (str): GPX sub-element.
            number (Union[int, float], optional): GPX sub-element value.
            precision (int, optional): Precision. Defaults to DEFAULT_PRECISION.

        Returns:
            Tuple[ET.Element, Union[ET.Element, None]]: GPX element and
                GPX sub-element (if not None).
        """
        sub_element_ = None
        if number is not None:
            sub_element_ = ET.SubElement(element, sub_element)
            if isinstance(number, int):
                sub_element_.text = str(number)
            elif isinstance(number, float):
                sub_element_.text = f"{number:.{precision}f}"
            else:
                logging.error("Invalid number type.")
        return element, sub_element_

    def add_subelement_time(
        self,
        element: ET.Element,
        sub_element: str,
        time: datetime,
        format_: str = DEFAULT_TIME_FORMAT,
    ) -> Tuple[ET.Element, Union[ET.Element, None]]:
        """
        Add sub-element to GPX element.

        Args:
            element (ET.Element): GPX element.
            sub_element (str): GPX sub-element.
            time (datetime, optional): GPX sub-element value.
            format (str, optional): Format. Defaults to DEFAULT_TIME_FORMAT.

        Returns:
            Tuple[ET.Element, Union[ET.Element, None]]: GPX element and
                GPX sub-element (if not None).
        """
        sub_element_ = None
        if time is not None:
            sub_element_ = ET.SubElement(element, sub_element)
            time_utc = time.astimezone(timezone.utc)  # Convert to UTC
            sub_element_.text = time_utc.strftime(format_)
        return element, sub_element_

    def check_xml_schemas(
        self, xml_schema: bool = False, xml_extensions_schemas: bool = False
    ) -> bool:
        """
        Check XML schemas after writting.

        Parameters
        ----------
        xml_schema : bool, optional
            Toggle XML schema verification, by default False.
        xml_extensions_schemas : bool, optional
            Toggle XML extensions schemas verification, by default False.

        Returns
        -------
        bool
            True if the written file follows all verified schemas.
        """
        # TODO Check for file_path
        # Check XML schema
        if xml_schema:
            if not self.gpx.check_xml_schema(self.file_path):
                logging.error("Invalid GPX file (does not follow XML schema).")
                return False

        # Check XML extension schemas
        if xml_extensions_schemas:
            if not self.gpx.check_xml_extensions_schemas(self.file_path):
                logging.error(
                    "Invalid GPX file (does not follow XML extensions schemas)."
                )
                return False

        return True



from datetime import datetime

from .extensions import Extensions
from .gpx_element import GpxElement
from .link import Link
from .person import Person


class Metadata(GpxElement):
    """
    metadataType element in GPX file.
    """

    fields = [
        "name",
        "desc",
        "author",
        "copyright",
        "link",
        "time",
        "keywords",
        "bounds",
        "extensions",
    ]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "metadata",
        name: str = None,
        desc: str = None,
        author: Person = None,
        copyright_: str = None,
        link: Link = None,
        time: datetime = None,
        keywords: str = None,
        bounds: str = None,
        extensions: Extensions = None,
    ) -> None:
        """
        Initialize Metadata instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "metadata"
        name : str, optional
            Name, by default None
        desc : str, optional
            Description, by default None
        author : Person, optional
            Author, by default None
        copyright : str, optional
            Copyright, by default None
        link : Link, optional
            Link, by default None
        time : datetime, optional
            Time, by default None
        keywords : str, optional
            Keywords, by default None
        bounds : str, optional
            Bounds, by default None
        extensions : Extensions, optional
            Extensions, by default None
        """
        self.tag: str = tag
        self.name: str = name
        self.desc: str = desc
        self.author: Person = author
        self.copyright: str = copyright_
        self.link: Link = link
        self.time: datetime = time
        self.keywords: str = keywords
        self.bounds: str = bounds
        self.extensions: str = extensions



import webbrowser
from typing import Optional, Tuple

import gmplot

from .plotter import Plotter


class GmplotPlotter(Plotter):
    """
    GPX plotter based on gmplot.
    """

    def plot(
        self,
        color: str = "#FFA800",
        start_stop_colors: Optional[Tuple[str, str]] = None,
        way_points_color: Optional[str] = None,
        scatter: bool = False,
        plot: bool = True,
        zoom: float = 10.0,
        title: Optional[str] = None,
        file_path: Optional[str] = None,
        browser: bool = True,
    ):
        """
        Plot GPX using gmplot.

        Args:
            color (str, optional): Track color. Defaults to "#FFA800".
            start_stop_colors (Optional[Tuple[str, str]], optional):
                Start and stop points colors. Defaults to None.
            way_points_color (Optional[str], optional): Way points
                color. Defaults to None.
            scatter (bool, optional): Scatter track points. Defaults
                to False.
            plot (bool, optional): Plot track points. Defaults to True.
            zoom (float, optional): Zoom. Defaults to 10.0.
            title (Optional[str], optional): Title of the plot.
                Defaults to None.
            file_path (Optional[str], optional): Path to save plot.
                Defaults to None.
            browser (bool, optional): Open the plot in the default
                web browser. Defaults to True.
        """
        # Create plotter
        c_lat, c_lon = self._gpx.center()
        map_ = gmplot.GoogleMapPlotter(c_lat, c_lon, zoom)

        # Create dataframe containing data from the GPX file
        gpx_df = self._gpx.to_pandas()

        # Scatter track points
        if scatter:
            map_.scatter(gpx_df["lat"], gpx_df["lon"], color, size=5, marker=False)
        if plot:
            map_.plot(gpx_df["lat"], gpx_df["lon"], color, edge_width=2.5)

        # Scatter start and stop points with different color
        if start_stop_colors:
            map_.scatter(
                [self._gpx.gpx.trk[0].trkseg[0].trkpt[0].lat],
                [self._gpx.gpx.trk[0].trkseg[0].trkpt[0].lon],
                start_stop_colors[0],
                size=5,
                marker=True,
            )
            map_.scatter(
                [self._gpx.gpx.trk[-1].trkseg[-1].trkpt[-1].lat],
                [self._gpx.gpx.trk[-1].trkseg[-1].trkpt[-1].lon],
                start_stop_colors[1],
                size=5,
                marker=True,
            )

        # Scatter way points with different color
        if way_points_color:
            for way_point in self._gpx.gpx.wpt:
                map_.scatter(
                    [way_point.lat],
                    [way_point.lon],
                    way_points_color,
                    size=5,
                    marker=True,
                )

        # Add title
        if title is not None:
            map_.text(c_lat, c_lon, self._gpx.name(), color="#FFFFFF")

        # Save map
        if file_path is None:
            file_path = self._gpx.file_path[:-4] + ".html"
        map_.draw(file_path)

        # Open map in web browser
        if browser:
            webbrowser.open(file_path)



from __future__ import annotations

import errno
import logging
import os
import warnings
from datetime import datetime
from math import degrees
from typing import Dict, List, Optional, Tuple, Type, Union
from zipfile import ZipFile

import pandas as pd

from ..gpx_elements import (
    Bounds,
    Copyright,
    Email,
    Extensions,
    Gpx,
    Link,
    Metadata,
    Person,
    Point,
    PointSegment,
    Route,
    Track,
    TrackSegment,
    WayPoint,
)
from ..parsers.fit_parser import FitParser
from ..parsers.gpx_parser import GPXParser
from ..parsers.kml_parser import KMLParser
from ..utils import EARTH_RADIUS
from ..writers.gpx_writer import GPXWriter
from ..writers.kml_writer import KMLWriter

# GPX = NewType("GPX", object)  # GPX forward declaration for type hint


class GPX:
    """
    High level GPX object.
    """

    TIME_RELATED_VALUES = ["time", "speed", "pace", "ascent_speed"]
    ELEVATION_RELATED_VALUES = ["ele", "ascent_rate", "ascent_speed"]

    def __init__(
        self,
        file_path: Optional[str] = None,
        xml_schema: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> None:
        """
        Initialise GPX instance.

        Parameters
        ----------
        file_path : Optional[str], optional
            Path to the file to parse, by default None
        xml_schema : bool, optional
            Toggle schema verification during parsing, by default True
        extensions_schemas : bool, optional
            Toggle extensions schema verificaton during parsing, by default False
        """
        # GPX file description
        self.file_path: str = None
        self.file_name: str = None

        # GPX file content
        self.gpx: Gpx = None
        self._ele_data: bool = False
        self._time_data: bool = False
        self._precisions: Dict = None
        self._time_format: str = None

        # Parsers
        self._gpx_parser: GPXParser = None
        self._kml_parser: KMLParser = None
        self._fit_parser: FitParser = None

        # Writers
        self._gpx_writer: GPXWriter = None
        self._kml_writer: KMLWriter = None

        # Utility attributes
        self._dataframe: pd.DataFrame = None

        # Empty GPX instance
        # For advanced use only
        if file_path is None:
            warnings.warn(
                "No file path provided, creating an empty GPX instance.", UserWarning
            )
            self.gpx = Gpx()
            # Writers
            self._gpx_writer = GPXWriter(self.gpx)
            self._kml_writer = KMLWriter(self.gpx)
        # Valid file
        elif isinstance(file_path, str) and os.path.isfile(file_path):
            self.file_path = file_path
            self.file_name = os.path.basename(file_path)

            # GPX
            if file_path.endswith(".gpx"):
                self._gpx_parser = GPXParser(
                    file_path, xml_schema, xml_extensions_schemas
                )
                self.gpx = self._gpx_parser.gpx
                self._ele_data = self._gpx_parser.ele_data
                self._time_data = self._gpx_parser.time_data
                self._precisions = self._gpx_parser.precisions
                self._time_format = self._gpx_parser.time_format

            # KML
            elif file_path.endswith(".kml"):
                self._kml_parser = KMLParser(
                    file_path, xml_schema, xml_extensions_schemas
                )
                self.gpx = self._kml_parser.gpx
                self._precisions = self._kml_parser.precisions
                self._time_format = self._kml_parser.time_format

            # KMZ
            elif file_path.endswith(".kmz"):
                kmz = ZipFile(file_path, "r")
                kmls = [
                    info.filename
                    for info in kmz.infolist()
                    if info.filename.endswith(".kml")
                ]
                if "doc.kml" not in kmls:
                    raise FileNotFoundError(
                        f"Unable to parse file: {file_path}"
                        "Expected to find doc.kml inside KMZ file."
                    )
                kml = kmz.open("doc.kml", "r").read()
                self._write_tmp_kml("tmp.kml", kml)
                self._kml_parser = KMLParser(
                    "tmp.kml", xml_schema, xml_extensions_schemas
                )
                self.gpx = self._kml_parser.gpx
                self._precisions = self._kml_parser.precisions
                self._time_format = self._kml_parser.time_format
                os.remove("tmp.kml")

            # FIT
            elif file_path.endswith(".fit"):
                self._fit_parser = FitParser(file_path)
                self.gpx = self._fit_parser.gpx
                self._precisions = self._fit_parser.precisions
                self._time_format = self._fit_parser.time_format

            # Invalid file or file path
            else:
                raise ValueError(
                    f"Unable to parse this type of file: {file_path}"
                    "Consider renaming your file with the proper file extension."
                )

            # Writers
            self._gpx_writer = GPXWriter(self.gpx, self._precisions, self._time_format)
            self._kml_writer = KMLWriter(
                self.gpx, precisions=self._precisions, time_format=self._time_format
            )

        # Invalid file path
        else:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)

    def _write_tmp_kml(self, path: str = "tmp.kml", kml_string: Optional[bytes] = None):
        """
        Write temproray .KML file in order to parse KMZ file.
        """
        # Open/create KML file
        try:
            f = open(path, "wb")
        except OSError:
            logging.exception("Could not open/read file: %s", path)
            raise
        # Write KML file
        with f:
            if kml_string is not None:
                f.write(kml_string)

    def __str__(self) -> str:
        return self._gpx_writer.gpx_to_string()

    def __repr__(self):
        return f"file_path = {self.file_path}\ngpx = {self.gpx}"

    ###############################################################################
    #### Schemas ##################################################################
    ###############################################################################

    def check_xml_schema(self) -> bool:
        """
        Check XML schema.

        Returns
        -------
        bool
            True if the file follows XML schemas.
        """
        return self.gpx.check_xml_schema(self.file_path)

    def check_xml_extensions_schemas(self) -> bool:
        """
        Check XML extension schemas.

        Returns
        -------
        bool
            True if the file follows XML schemas.
        """
        return self.gpx.check_xml_extensions_schemas(self.file_path)

    ###############################################################################
    #### Name #####################################################################
    ###############################################################################

    def name(self) -> str:
        """
        Return activity name.

        Returns
        -------
        str
            Activity name.
        """
        return self.gpx.name()

    def set_name(self, new_name: str) -> None:
        """
        Set name.

        Parameters
        ----------
        new_name : str
            New name.
        """
        self.gpx.set_name(new_name)

    ###############################################################################
    #### Points ###################################################################
    ###############################################################################

    def nb_points(self) -> int:
        """
        Return the number of points in the GPX.

        Returns
        -------
        int
            Number of points in the GPX.
        """
        return self.gpx.nb_points()

    def bounds(self) -> Tuple[float, float, float, float]:
        """
        Find minimum and maximum latitude and longitude.

        Returns
        -------
        Tuple[float, float, float, float]
            Min latitude, min longitude, max latitude, max longitude.
        """
        return self.gpx.bounds()

    def center(self) -> Tuple[float, float]:
        """
        Return the coordinates of the center point.

        Returns
        -------
        Tuple[float, float]
            Latitude and longitude of the center point.
        """
        return self.gpx.center()

    def first_point(self) -> WayPoint:
        """
        Return GPX first point.

        Returns
        -------
        WayPoint
            First point.
        """
        return self.gpx.first_point()

    def last_point(self) -> WayPoint:
        """
        Return GPX last point.

        Returns
        -------
        WayPoint
            Last point.
        """
        return self.gpx.last_point()

    def extreme_points(self) -> Tuple[WayPoint, WayPoint, WayPoint, WayPoint]:
        """
        Find extreme points in track, i.e.: points with lowest and highest latitude and longitude.

        Returns
        -------
        Tuple[WayPoint, WayPoint, WayPoint, WayPoint]
            Min latitude point, min longitude point, max latitude point, max longitude point
        """
        return self.gpx.extreme_points()

    ###############################################################################
    #### Distance and Elevation ###################################################
    ###############################################################################

    def distance(self) -> float:
        """
        Returns the distance (meters) of tracks contained in the GPX.

        Returns
        -------
        float
            Distance (meters).
        """
        return self.gpx.distance()

    def ascent(self) -> float:
        """
        Returns the ascent (meters) of tracks contained in the GPX.

        Returns
        -------
        float
            Ascent (meters).
        """
        return self.gpx.ascent()

    def descent(self) -> float:
        """
        Returns the descent (meters) of tracks contained in the GPX.

        Returns
        -------
        float
            Descent (meters).
        """
        return self.gpx.descent()

    def min_elevation(self) -> float:
        """
        Returns the minimum elevation (meters) in tracks contained in the GPX.

        Returns
        -------
        float
            Minimum elevation (meters).
        """
        return self.gpx.min_elevation()

    def max_elevation(self) -> float:
        """
        Returns the maximum elevation (meters) in tracks contained in the GPX.

        Returns
        -------
        float
            Maximum elevation (meters).
        """
        return self.gpx.max_elevation()

    def compute_points_ascent_rate(self) -> None:
        """
        Compute ascent rate at each point.
        """
        self.gpx.compute_points_ascent_rate()

    def min_ascent_rate(self) -> float:
        """
        Return activity minimum ascent rate.

        Returns
        -------
        float
            Minimum ascent rate.
        """
        return self.gpx.min_ascent_rate()

    def max_ascent_rate(self) -> float:
        """
        Return activity maximum ascent rate.

        Returns
        -------
        float
            Maximum ascent rate.
        """
        return self.gpx.max_ascent_rate()

    ###############################################################################
    #### Time #####################################################################
    ###############################################################################

    def start_time(self) -> datetime:
        """
        Return the activity start time.

        Returns
        -------
        datetime
            Start time.
        """
        return self.gpx.start_time()

    def stop_time(self) -> datetime:
        """
        Return the activity stop time.

        Returns
        -------
        datetime
            Stop time.
        """
        return self.gpx.stop_time()

    def total_elapsed_time(self) -> datetime:
        """
        Return the total elapsed time during the activity.

        Returns
        -------
        datetime
            Total elapsed time.
        """
        return self.gpx.total_elapsed_time()

    def stopped_time(self) -> datetime:
        """
        Return the stopped time during the activity.

        Returns
        -------
        datetime
            Stopped time.
        """
        return self.gpx.stopped_time()

    def moving_time(self) -> datetime:
        """
        Return the moving time during the activity.

        Returns
        -------
        datetime
            Moving time.
        """
        return self.gpx.moving_time()

    ###############################################################################
    #### Speed and Pace ###########################################################
    ###############################################################################

    def avg_speed(self) -> float:
        """
        Return average speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Average speed (kilometres per hour).
        """
        return self.gpx.avg_speed()

    def avg_moving_speed(self) -> float:
        """
        Return average moving speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Average moving speed (kilometres per hour).
        """
        return self.gpx.avg_moving_speed()

    def compute_points_speed(self) -> None:
        """
        Compute speed (kilometres per hour) at each track point.
        """
        self.gpx.compute_points_speed()

    def min_speed(self) -> float:
        """
        Return the minimum speed during the activity.

        Returns
        -------
        float
            Minimum speed.
        """
        return self.gpx.min_speed()

    def max_speed(self) -> float:
        """
        Return the maximum speed during the activity.

        Returns
        -------
        float
            Maximum speed.
        """
        return self.gpx.max_speed()

    def avg_pace(self) -> float:
        """
        Return average pace (minutes per kilometer) during the activity.

        Returns
        -------
        float
            Average pace (minutes per kilometer).
        """
        return self.gpx.avg_pace()

    def avg_moving_pace(self) -> float:
        """
        Return average moving pace (minutes per kilometer) during the activity.

        Returns
        -------
        float
            Average moving pace (minutes per kilometer).
        """
        return self.gpx.avg_moving_pace()

    def compute_points_pace(self) -> None:
        """
        Compute pace at each track point.
        """
        self.gpx.compute_points_pace()

    def min_pace(self) -> float:
        """
        Return the minimum pace during the activity.

        Returns
        -------
        float
            Minimum pace.
        """
        return self.gpx.min_pace()

    def max_pace(self) -> float:
        """
        Return the maximum pace during the activity.

        Returns
        -------
        float
            Maximum pace.
        """
        return self.gpx.max_pace()

    def compute_points_ascent_speed(self) -> None:
        """
        Compute ascent speed (kilometres per hour) at each track point.
        """
        self.gpx.compute_points_ascent_speed()

    def min_ascent_speed(self) -> float:
        """
        Return the minimum ascent speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Minimum ascent speed.
        """
        return self.gpx.min_ascent_speed()

    def max_ascent_speed(self) -> float:
        """
        Return the maximum ascent speed (kilometres per hour) during the activity.

        Returns
        -------
        float
            Maximum ascent speed.
        """
        return self.gpx.max_ascent_speed()

    ###############################################################################
    #### Data Removal #############################################################
    ###############################################################################

    def remove_metadata(self):
        """
        Remove metadata.
        """
        self.gpx.remove_metadata()

    def remove_elevation(self):
        """
        Remove elevation data.
        """
        self.gpx.remove_elevation()

    def remove_time(self):
        """
        Remove time data.
        """
        self.gpx.remove_time()

    def remove_extensions(self):
        """
        Remove extensions data.
        """
        self.gpx.remove_extensions()

    ###############################################################################
    #### Error Correction #########################################################
    ###############################################################################

    def remove_gps_errors(self):
        """
        Remove GPS errors.
        """
        self.gpx.remove_gps_errors()

    def remove_close_points(self, min_dist: float = 1, max_dist: float = 10):
        """
        Remove points that are to close together.

        Parameters
        ----------
        min_dist : float, optional
            Minimal distance between two points, by default 1
        max_dist : float, optional
            Maximal distance between two points, by default 10
        """
        self.gpx.remove_close_points(min_dist, max_dist)

    ###############################################################################
    #### Simplification ###########################################################
    ###############################################################################

    def simplify(self, tolerance: float = 2):
        """
        Simplify tracks using Ramer-Douglas-Peucker algorithm.

        Parameters
        ----------
        tolerance : float, optional
            Tolerance (meters). Corresponds to the minimum distance between the
            point and the track before the point is removed, by default 2
        """
        epsilon = degrees(tolerance / EARTH_RADIUS)
        self.gpx.simplify(epsilon)

    ###############################################################################
    #### Merge ####################################################################
    ###############################################################################

    @staticmethod
    def merge(gpx_1: GPX, gpx_2: GPX) -> GPX:
        """
        Merge GPX objects in a new instance.

        Parameters
        ----------
        gpx_1 : GPX
            First GPX object
        gpx_2 : GPX
            Second GPX object

        Returns
        -------
        GPX
            Merged GPX (new instance)
        """
        topo = [
            "http://www.topografix.com/GPX/1/1",
            "http://www.topografix.com/GPX/1/1/gpx.xsd",
        ]

        # Create new GPX instance
        merged_gpx = GPX()

        # Fill new GPX instance
        merged_gpx.gpx.tag = "gpx"
        merged_gpx.gpx.xmlns = "http://www.topografix.com/GPX/1/1"
        merged_gpx.gpx.xsi_schema_location = list(
            set(topo + gpx_1.gpx.xsi_schema_location + gpx_2.gpx.xsi_schema_location)
        )
        merged_gpx.gpx.version = "1.1"
        merged_gpx.gpx.creator = "ezGPX"
        merged_gpx.gpx.metadata = (
            gpx_2.gpx.metadata if gpx_1.gpx.metadata is None else gpx_1.gpx.metadata
        )
        merged_gpx.gpx.wpt = gpx_1.gpx.wpt + gpx_2.gpx.wpt
        merged_gpx.gpx.rte = gpx_1.gpx.rte + gpx_2.gpx.rte
        merged_gpx.gpx.trk = gpx_1.gpx.trk + gpx_2.gpx.trk
        merged_gpx.gpx.extensions = Extensions(
            "extensions", gpx_1.gpx.metadata | gpx_2.gpx.metadata
        )

        # Return new GPX instance
        return merged_gpx

    ###############################################################################
    #### Exports ##################################################################
    ###############################################################################

    def to_dict(
        self,
        values: List[str] = None,
        orient: str = "dict",
        into: Type[dict] = dict,
        index: bool = True,
    ) -> Dict:
        """
        Convert GPX object to dictionary.
        Pandas.DataFrame.to_dict documentation: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html

        Parameters
        ----------
        values : List[str], optional
            values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"
        orient : str, optional
            Same as in Pandas.DataFrame.to_dict, by default "dict"
        into : Type[dict], optional
            Same as in Pandas.DataFrame.to_dict, by default dict
        index : bool, optional
            Same as in Pandas.DataFrame.to_dict, by default True

        Returns
        -------
        Dict
            Return a dictionary representing the GPX. The resulting
            transformation depends on the `orient` parameter.
        """
        return self.gpx.to_dict(values, orient, into, index)

    def to_pandas(self, values: List[str] = None) -> pd.DataFrame:
        """
        Convert GPX object to Pandas Dataframe.
        Missing values are filled with default values (0 for elevation, empty string for time).

        Parameters
        ----------
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"

        Returns
        -------
        pd.DataFrame
            Dataframe containing data from GPX.
        """
        # Disable time related values if no time data available
        if not self._time_data:
            if any(v in GPX.TIME_RELATED_VALUES for v in values):
                warnings.warn(
                    f"Trying to create dataframe from GPX file {self.file_path} which does not contain time data"
                    "Time related values (time, speed, pace, ascent speed) will not be present in the dataframe.",
                    UserWarning,
                )
            for v in GPX.TIME_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        # Disable elevation related values if no elevation data available
        if not self._ele_data:
            if any(v in GPX.ELEVATION_RELATED_VALUES for v in values):
                warnings.warn(
                    f"Trying to create dataframe from GPX file {self.file_path} which does not contain elevation data"
                    "Time related values (elevation, ascent rate, ascent speed) will not be present in the dataframe.",
                    UserWarning,
                )
            for v in GPX.ELEVATION_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        return self.gpx.to_pandas(values)

    def to_polars(self, values: List[str] = None) -> pd.DataFrame:
        """
        Convert GPX object to Polars Dataframe.
        Missing values are filled with default values (0 for elevation, empty string for time).

        Parameters
        ----------
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"

        Returns
        -------
        pd.DataFrame
            Dataframe containing data from GPX.
        """
        # Disable time related values if no time data available
        if not self._time_data:
            if any(v in GPX.TIME_RELATED_VALUES for v in values):
                warnings.warn(
                    f"Trying to create dataframe from GPX file {self.file_path} which does not contain time data"
                    "Time related values (time, speed, pace, ascent speed) will not be present in the dataframe.",
                    UserWarning,
                )
            for v in GPX.TIME_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        # Disable elevation related values if no elevation data available
        if not self._ele_data:
            if any(v in GPX.ELEVATION_RELATED_VALUES for v in values):
                warnings.warn(
                    f"Trying to create dataframe from GPX file {self.file_path} which does not contain elevation data"
                    "Time related values (elevation, ascent rate, ascent speed) will not be present in the dataframe.",
                    UserWarning,
                )
            for v in GPX.ELEVATION_RELATED_VALUES:
                if v in values:
                    values.remove(v)

        return self.gpx.to_polars(values)

    def to_csv(
        self,
        path: str = None,
        values: List[str] = None,
        sep: str = ",",
        header: bool = True,
        index: bool = False,
    ) -> Union[str, None]:  # TODO select pandas vs polars
        """
        Write the GPX object track coordinates to a .csv file.

        Parameters
        ----------
        path : str, optional
            Path to the .csv file, by default None
        values : List[str], optional
            List of values to write, by default None
            Supported values: "lat", "lon", "ele", "time", "speed", "pace",
            "ascent_rate", "ascent_speed", "distance_from_start"
        sep : str, optional
            Separator, by default ","
        header : bool, optional
            Toggle header, by default True
        index : bool, optional
             Toggle index, by default False

        Returns
        -------
        str
            CSV like string if path is set to None.
        """
        return self.gpx.to_csv(path, values, sep, header, index)

    def to_gpx(
        self,
        path: str,
        properties: bool = True,
        bounds_fields: Optional[List[str]] = None,
        copyright_fields: Optional[List[str]] = None,
        email_fields: Optional[List[str]] = None,
        extensions_fields: Optional[Dict] = None,
        gpx_fields: Optional[List[str]] = None,
        link_fields: Optional[List[str]] = None,
        metadata_fields: Optional[List[str]] = None,
        person_fields: Optional[List[str]] = None,
        point_segment_fields: Optional[List[str]] = None,
        point_fields: Optional[List[str]] = None,
        route_fields: Optional[List[str]] = None,
        track_segment_fields: Optional[List[str]] = None,
        track_fields: Optional[List[str]] = None,
        way_point_fields: Optional[List[str]] = None,
        track_point_fields: Optional[List[str]] = None,
        mandatory_fields: bool = True,
        xml_schema: bool = True,
        xml_extensions_schemas: bool = False,
    ) -> bool:
        """
        Write the GPX object to a .gpx file.

        Parameters
        ----------
        path : str
            Path to the new .gpx file.
        xml_schema : bool, optional
            Toggle schema verification after writting, by default True
        xml_extensions_schemas : bool, optional
            Toggle extensions schema verificaton after writing. Requires internet connection and is not guaranted to work, by default False

        Returns
        -------
        bool
            Return False if written file does not follow checked schemas. Return True otherwise.
        """
        bounds_fields = bounds_fields if bounds_fields is not None else Bounds.fields
        copyright_fields = (
            copyright_fields if copyright_fields is not None else Copyright.fields
        )
        email_fields = email_fields if email_fields is not None else Email.fields
        default_extensions_fields = (
            self._gpx_parser.extensions_fields if self._gpx_parser is not None else {}
        )
        extensions_fields = (
            extensions_fields
            if extensions_fields is not None
            else default_extensions_fields
        )
        gpx_fields = gpx_fields if gpx_fields is not None else Gpx.fields
        link_fields = link_fields if link_fields is not None else Link.fields
        metadata_fields = (
            metadata_fields if metadata_fields is not None else Metadata.fields
        )
        person_fields = person_fields if person_fields is not None else Person.fields
        point_segment_fields = (
            point_segment_fields
            if point_segment_fields is not None
            else PointSegment.fields
        )
        point_fields = point_fields if point_fields is not None else Point.fields
        route_fields = route_fields if route_fields is not None else Route.fields
        track_segment_fields = (
            track_segment_fields
            if track_segment_fields is not None
            else TrackSegment.fields
        )
        track_fields = track_fields if track_fields is not None else Track.fields
        way_point_fields = (
            way_point_fields if way_point_fields is not None else WayPoint.fields
        )
        track_point_fields = (
            track_point_fields if track_point_fields is not None else WayPoint.fields
        )
        return self._gpx_writer.write(
            file_path=path,
            properties=properties,
            bounds_fields=bounds_fields,
            copyright_fields=copyright_fields,
            email_fields=email_fields,
            extensions_fields=extensions_fields,
            gpx_fields=gpx_fields,
            link_fields=link_fields,
            metadata_fields=metadata_fields,
            person_fields=person_fields,
            point_segment_fields=point_segment_fields,
            point_fields=point_fields,
            route_fields=route_fields,
            track_segment_fields=track_segment_fields,
            track_fields=track_fields,
            way_point_fields=way_point_fields,
            track_point_fields=track_point_fields,
            mandatory_fields=mandatory_fields,
            xml_schema=xml_schema,
            xml_extensions_schemas=xml_extensions_schemas,
        )

    def to_kml(
        self,
        path: str,
        styles: Optional[List[Tuple[str, Dict]]] = None,
        xml_schema: bool = True,
    ) -> bool:
        """
        Write the GPX object to a .kml file.

        Parameters
        ----------
        path : str
            Path to the .gpx file.
        styles : List[Tuple[str, Dict]], optional
            List of (style_id, style) tuples, by default None
        xml_schema : bool, optional
            Toggle schema verification after writting, by default True

        Returns
        -------
        bool
            Return False if written file does not follow checked schemas. Return True otherwise.
        """
        return self._kml_writer.write(path, styles, xml_schema)



import webbrowser
from typing import Optional, Tuple

import folium
from folium.features import DivIcon
from folium.plugins import MiniMap

from .plotter import Plotter


class FoliumPlotter(Plotter):
    """
    GPX plotter based on folium.
    """

    def plot(
        self,
        tiles: str = "OpenStreetMap",  # "OpenStreetMap", "Stamen Terrain", "Stamen Toner"
        color: str = "#FFA800",
        start_stop_colors: Optional[Tuple[str, str]] = None,
        way_points_color: Optional[str] = None,
        minimap: bool = False,
        coord_popup: bool = False,
        title: Optional[str] = None,
        zoom: float = 12.0,
        file_path: Optional[str] = None,
        browser: bool = True,
    ):
        """
        Plot GPX using folium.

        Args:
            tiles (str, optional): Map tiles. Supported tiles:
                "OpenStreetMap", "Stamen Terrain", "Stamen Toner".
                Defaults to "OpenStreetMap".
            start_stop_colors (Optional[Tuple[str, str]], optional):
                Start and stop points colors. Defaults to None.
            way_points_color (Optional[str], optional): Way points
                color. Defaults to None.
            minimap (bool, optional): Add minimap. Defaults to False.
            coord_popup (bool, optional): Add coordinates pop-up when
                clicking on the map. Defaults to False.
            title (Optional[str], optional): Title of the plot.
                Defaults to None.
            zoom (float, optional): Zoom. Defaults to 12.0.
            file_path (Optional[str], optional): Path to save plot.
                Defaults to None.
            browser (bool, optional): Open the plot in the default web
                browser. Defaults to True.
        """
        # Create map
        center_lat, center_lon = self._gpx.center()
        m = folium.Map(location=[center_lat, center_lon], zoom_start=zoom, tiles=tiles)

        # Plot track points
        gpx_df = self._gpx.to_pandas()
        gpx_df["coordinates"] = list(zip(gpx_df.lat, gpx_df.lon))
        folium.PolyLine(
            gpx_df["coordinates"], tooltip=self._gpx.name(), color=color
        ).add_to(m)

        # Scatter start and stop points with different color
        if start_stop_colors:
            folium.Marker(
                [
                    self._gpx.gpx.trk[0].trkseg[0].trkpt[0].lat,
                    self._gpx.gpx.trk[0].trkseg[0].trkpt[0].lon,
                ],
                popup="<b>Start</b>",
                tooltip="Start",
                icon=folium.Icon(color=start_stop_colors[0]),
            ).add_to(m)
            folium.Marker(
                [
                    self._gpx.gpx.trk[-1].trkseg[-1].trkpt[-1].lat,
                    self._gpx.gpx.trk[-1].trkseg[-1].trkpt[-1].lon,
                ],
                popup="<b>Stop</b>",
                tooltip="Stop",
                icon=folium.Icon(color=start_stop_colors[1]),
            ).add_to(m)

        # Scatter way points with different color
        if way_points_color:
            for way_point in self._gpx.gpx.wpt:
                folium.Marker(
                    [way_point.lat, way_point.lon],
                    popup="<i>Way point</i>",
                    tooltip="Way point",
                    icon=folium.Icon(icon="info-sign", color=way_points_color),
                ).add_to(m)

        # Add minimap
        if minimap:
            minimap = MiniMap(toggle_display=True)
            minimap.add_to(m)

        # Add latitude-longitude pop-up
        if coord_popup:
            m.add_child(folium.LatLngPopup())

        # Title
        if title is not None:
            folium.map.Marker(
                [center_lat, center_lon],
                icon=DivIcon(
                    icon_size=(250, 36),
                    icon_anchor=(0, 0),
                    html=f'<div style="font-size: 20pt">{title}</div>',
                ),
            ).add_to(m)

        # Save map
        if file_path is None:
            file_path = self._gpx.file_path[:-4] + ".html"
        m.save(file_path)

        # Open map in web browser
        if browser:
            webbrowser.open(file_path)



from .gpx_element import GpxElement


class Bounds(GpxElement):
    """
    boundsType element in GPX file.
    """

    fields = ["minlat", "minlon", "maxlat", "maxlon"]
    mandatory_fields = ["minlat", "minlon", "maxlat", "maxlon"]

    def __init__(
        self,
        tag: str = "bounds",
        minlat: float = None,
        minlon: float = None,
        maxlat: float = None,
        maxlon: float = None,
    ) -> None:
        """
        Initialize Bounds instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "bounds"
        minlat : float, optional
            Minimal latitude, by default None
        minlon : float, optional
            Minimal longitude, by default None
        maxlat : float, optional
            Maximal latitude, by default None
        maxlon : float, optional
            Maximal longitude, by default None
        """
        self.tag: str = tag
        self.minlat: float = minlat
        self.minlon: float = minlon
        self.maxlat: float = maxlat
        self.maxlon: float = maxlon



from typing import List

from .extensions import Extensions
from .gpx_element import GpxElement
from .link import Link
from .way_point import WayPoint


class Route(GpxElement):
    """
    rteType element in GPX file.
    """

    fields = [
        "name",
        "cmt",
        "desc",
        "src",
        "link",
        "number",
        "type",
        "extensions",
        "rtept",
    ]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "rte",
        name: str = None,
        cmt: str = None,
        desc: str = None,
        src: str = None,
        link: Link = None,
        number: int = None,  # non negative integer
        type_: str = None,
        extensions: Extensions = None,
        rtept: List[WayPoint] = None,
    ) -> None:
        """
        Initialize Route instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "rte"
        name : str, optional
            Name, by default None
        cmt : str, optional
            Comment, by default None
        desc : str, optional
            Description, by default None
        src : str, optional
            Source, by default None
        link : Link, optional
            Link, by default None
        number : int, optional
            Number, by default None
        extensions : Extensions, optional
            Extensions, by default None
        rtept : List[WayPoint], optional
            Route points, by default None
        """
        self.tag: str = tag
        self.name: str = name
        self.cmt: str = cmt
        self.desc: str = desc
        self.src: str = src
        self.link: Link = link
        self.number: int = number
        self.type: str = type_
        self.extensions: Extensions = extensions
        self.rtept: List[WayPoint] = [] if rtept is None else rtept



import filecmp
import os
import sys
from shutil import rmtree

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

file_folder = os.path.dirname(__file__)
parent_folder = os.path.realpath(os.path.dirname(file_folder))  # ie: ezGPX

os.chdir(file_folder)
sys.path.append(parent_folder + "/ezgpx")

from ezgpx import GPX

FILES_DIRECTORY = "test_files/files/"
REFERENCE_FILES_DIRECTORY = "test_files/reference_files/"


class TestGPX:

    # ==== Init ===============================================================#

    def test_init(self):
        # Create temporary folder
        rmtree("tmp", True)
        os.makedirs(os.path.dirname(__file__) + "/tmp")

    # ==== Parsing/Writing ====================================================#

    def test_parse(self):
        # Parse GPX files
        gpx = GPX(
            os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"),
            xml_schema=True,
            xml_extensions_schemas=False,
        )
        invalid_gpx = GPX(
            os.path.join(FILES_DIRECTORY, "invalid_schema.gpx"),
            xml_schema=False,
            xml_extensions_schemas=False,
        )

    # ==== Check Schemas ======================================================#

    def test_check_schemas(self):
        # Parse GPX Files
        gpx = GPX(
            os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"),
            xml_schema=False,
            xml_extensions_schemas=False,
        )
        invalid_gpx = GPX(
            os.path.join(FILES_DIRECTORY, "invalid_schema.gpx"),
            xml_schema=False,
            xml_extensions_schemas=False,
        )
        # Tests
        assert gpx.check_xml_schema() is True
        assert invalid_gpx.check_xml_schema() is False

    # ==== Properties =========================================================#

    def test_file_name(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.file_name == "strava_run_1.gpx"

    def test_name(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.name() == "Dérouillage habituel 💥"

    def test_set_name(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        new_name = "test"
        gpx.set_name(new_name)
        assert gpx.name() == new_name

    def test_nb_points(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.nb_points() == 939

    @pytest.mark.skip(reason="nothing to test")
    def test_first_point(self):
        pass

    @pytest.mark.skip(reason="nothing to test")
    def test_last_point(self):
        pass

    def test_bounds(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.bounds() == (44.032965, 4.444134, 44.047778, 4.486607)

    def test_center(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.center() == (44.0403715, 4.465370500000001)

    def test_distance(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.distance() == 10922.788757238777

    def test_ascent(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.ascent() == 225.29999999999995

    def test_descent(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.descent() == 224.79999999999987

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_ascent_rate(self):
        pass

    def test_min_ascent_rate(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.min_ascent_rate() == -38.54138626353898

    def test_max_ascent_rate(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.max_ascent_rate() == 51.2919872933083

    def test_min_elevation(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.min_elevation() == 98.5

    def test_max_elevation(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.max_elevation() == 235.6

    @pytest.mark.skip(reason="time related test")
    def test_start_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_stop_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_total_elapsed_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_stopped_time(self):
        pass

    @pytest.mark.skip(reason="time related test")
    def test_moving_time(self):
        pass

    def test_avg_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.avg_speed() == 10.66505004775145

    def test_avg_moving_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.avg_moving_speed() == 10.974613320139435

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_speed(self):
        pass

    def test_min_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.min_speed() == 0.0

    def test_max_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.max_speed() == 19.69510525631602

    def test_avg_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.avg_pace() == 5.625852643105975

    def test_avg_moving_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.avg_moving_pace() == 5.467163010645161

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_pace(self):
        pass

    def test_min_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.min_pace() == 3.046442210851278

    def test_max_pace(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.max_pace() == 1041.2956304834424

    @pytest.mark.skip(reason="nothing to test")
    def test_compute_points_ascent_speed(self):
        pass

    def test_min_ascent_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.min_ascent_speed() == -3.3599999999999794

    def test_max_ascent_speed(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Test
        assert gpx.max_ascent_speed() == 2.1599999999999797

    # ==== Modifications ======================================================#

    # ==== Conversion and Saving ==============================================#

    def test_to_pandas(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        df = gpx.to_pandas(values=["lat", "lon", "ele", "time"])
        reference_df = pd.read_csv(
            os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1.csv")
        )
        # Test
        assert reference_df.equals(df)

    def test_to_gpx(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        gpx.to_gpx("tmp/strava_run_1_test.gpx", xml_schema=False)
        # Test
        assert filecmp.cmp(
            "tmp/strava_run_1_test.gpx",
            os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1.gpx"),
            False,
        )

    def test_to_kml(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        gpx.to_kml("tmp/strava_run_1_test.kml", styles=None, xml_schema=False)
        # Test
        assert filecmp.cmp(
            "tmp/strava_run_1_test.kml",
            os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1.kml"),
            False,
        )

    def test_to_csv(self):
        # Parse GPX Files
        gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        gpx.to_csv("tmp/strava_run_1_test.csv", values=["lat", "lon", "ele", "time"])
        # Test
        assert filecmp.cmp(
            "tmp/strava_run_1_test.csv",
            os.path.join(REFERENCE_FILES_DIRECTORY, "strava_run_1.csv"),
            False,
        )

    # ==== Plots ==============================================================#

    # def _test_matplotlib_plot_1(self):
    #     # Plot
    #     self.gpx.matplotlib_plot(
    #         start_stop_colors=None, color="#ffffff", title="Track",
    #         file_path="tmp/matplotlib_strava_run_1.png")
    #     # Load images
    #     test_img = plt.imread("tmp/matplotlib_strava_run_1.png")
    #     ref_img = plt.imread(
    #         os.path.join(REFERENCE_FILES_DIRECTORY,
    #                      "matplotlib_strava_run_1.png"))
    #     # Compare images
    #     return np.array_equal(test_img, ref_img)

    # def _test_matplotlib_plot_2(self):
    #     # Plot
    #     self.gpx.matplotlib_plot(
    #         start_stop_colors=("#00FF00", "#FF0000"), color="#ffffff",
    #         title="Track", file_path="tmp/matplotlib_strava_run_1_start_stop.png")
    #     # Load images
    #     test_img = plt.imread("tmp/matplotlib_strava_run_1_start_stop.png")
    #     ref_img = plt.imread(
    #         os.path.join(REFERENCE_FILES_DIRECTORY,
    #                      "matplotlib_strava_run_1_start_stop.png"))

    #     # Compare images
    #     return np.array_equal(test_img, ref_img)

    # def _test_matplotlib_plot_3(self):
    #     # Plot
    #     self.gpx.matplotlib_plot(
    #         start_stop_colors=None, color=True, title="Track",
    #         file_path="tmp/matplotlib_strava_run_1_elevation.png")
    #     # Load images
    #     test_img = plt.imread("tmp/matplotlib_strava_run_1_elevation.png")
    #     ref_img = plt.imread(
    #         os.path.join(REFERENCE_FILES_DIRECTORY,
    #                      "matplotlib_strava_run_1_elevation.png"))
    #     # Compare images
    #     return np.array_equal(test_img, ref_img)

    # def _test_matplotlib_plot_4(self):
    #     # Plot
    #     self.gpx.matplotlib_plot(
    #         start_stop_colors=("#00FF00", "#FF0000"), color=True, title="Track",
    #         file_path="tmp/matplotlib_strava_run_1_start_stop_elevation.png")
    #     # Load images
    #     test_img = plt.imread("tmp/matplotlib_strava_run_1_start_stop_elevation.png")
    #     ref_img = plt.imread(
    #         os.path.join(REFERENCE_FILES_DIRECTORY,
    #                      "matplotlib_strava_run_1_start_stop_elevation.png"))
    #     # Compare images
    #     return np.array_equal(test_img, ref_img)

    # @pytest.mark.skip(reason="not ready")
    # def test_matplotlib_plot(self):
    #     # Parse GPX file
    #     self.gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
    #     # Tests
    #     assert self._test_matplotlib_plot_1()
    #     assert self._test_matplotlib_plot_2()
    #     assert self._test_matplotlib_plot_3()
    #     assert self._test_matplotlib_plot_4()
    #     # self._test_matplotlib_plot_1()
    #     # self._test_matplotlib_plot_2()
    #     # self._test_matplotlib_plot_3()
    #     # self._test_matplotlib_plot_4()

    # def _test_folium_plot_1(self):
    #     # Plot
    #     self.gpx.folium_plot(tiles="openStreetMap",
    #                          color="#110000",
    #                          start_stop_colors=None,
    #                          way_points_color=None,
    #                          minimap=False,
    #                          coord_popup=False,
    #                          title=None,
    #                          zoom=12,
    #                          file_path="tmp/folium_strava_run_1.html",
    #                          open=False)
    #     # Compare files
    #     return filecmp.cmp("tmp/folium_strava_run_1.html",
    #                        os.path.join(REFERENCE_FILES_DIRECTORY,
    #                                     "folium_strava_run_1.html"), False)

    # @pytest.mark.skip(reason="not ready")
    # def test_folium_plot(self):
    #     self.test_init() # For developping purpose only (using: pytest test_GPX.py::TestGPX::test_folium_plot)
    #     # Parse GPX file
    #     self.gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
    #     # Tests
    #     assert self._test_folium_plot_1()
    #     # assert self._test_folium_plot_2())
    #     # assert self._test_folium_plot_3())
    #     # assert self._test_folium_plot_4())
    #     # self._test_folium_plot_1()
    #     # self._test_folium_plot_2()
    #     # self._test_folium_plot_3()
    #     # self._test_folium_plot_4()

    # ==== Destroy ============================================================#

    def test_destroy(self, remove_tmp: bool = True):
        # Remove temporary folder
        if remove_tmp:
            rmtree("tmp", True)

    # ==== Test ===============================================================#

    @pytest.mark.skip(
        reason="test"
    )  # https://docs.pytest.org/en/7.3.x/how-to/skipping.html
    def test_test(self, remove_tmp: bool = True):
        # Create temporary folder
        rmtree("tmp", True)
        os.makedirs(os.path.dirname(__file__) + "/tmp")
        # Parse GPX file
        self.gpx = GPX(os.path.join(FILES_DIRECTORY, "strava_run_1.gpx"))
        # Remove temporary folder
        if remove_tmp:
            rmtree("tmp")



class GpxElement:
    """
    Base class for element in GPX file.
    Implements dunders functions.
    """

    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        s = "("
        for a in self.mandatory_fields:
            s += f"{getattr(self, a)}, "
        s = s[:-2] + ")"
        return s

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}[{self.tag}]("
        for a in self.mandatory_fields:
            s += f"{getattr(self, a)}, "
        s = s[:-2] + ")"
        return s



import logging
from typing import Dict, Optional, Union

from ..gpx_elements import Gpx

DEFAULT_PRECISION = 10
DEFAULT_PRECISION_DICT = {
    "lat_lon": DEFAULT_PRECISION,
    "elevation": DEFAULT_PRECISION,
    "distance": DEFAULT_PRECISION,
    "duration": DEFAULT_PRECISION,
    "speed": DEFAULT_PRECISION,
    "rate": DEFAULT_PRECISION,
    "default": DEFAULT_PRECISION,
}
POSSIBLE_TIME_FORMATS = ["%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S.%fZ"]
DEFAULT_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Parser:
    """
    File parser.
    """

    def __init__(
        self, file_path: Optional[str] = None, name_spaces: Dict = None
    ) -> None:
        """
        Initialize Parser instance.

        Args:
            file_path (str, optional): Path to the file to parse. Defaults to None.
        """
        self.file_path: str = file_path

        self.ele_data: bool = False
        self.time_data: bool = False
        self.precisions: dict = {
            "lat_lon": DEFAULT_PRECISION,
            "elevation": DEFAULT_PRECISION,
            "distance": DEFAULT_PRECISION,
            "duration": DEFAULT_PRECISION,
            "speed": DEFAULT_PRECISION,
            "rate": DEFAULT_PRECISION,
            "default": DEFAULT_PRECISION,
        }
        self.time_format: str = DEFAULT_TIME_FORMAT

        self.gpx: Gpx = Gpx(xmlns=name_spaces)

    def find_precision(self, number: Union[int, float, str]) -> int:
        """
        Find decimal precision of a given number.

        Parameters
        ----------
        number : Union[int, float, str]
            Number.

        Returns
        -------
        int
            Decimal precision.
        """
        if number is None:
            return DEFAULT_PRECISION

        if isinstance(number, int):
            return 0
        elif isinstance(number, float):
            number = str(number)
        elif isinstance(number, str):
            try:
                float(number)
            except OSError as err:
                logging.exception("OS error: %s", err)
            except ValueError:
                logging.exception(
                    "Could not convert data (%s) to a floating" "point value.", number
                )
            except Exception as err:
                logging.exception(
                    "Unexpected %s, %s." "Unable to find precision of number: %s",
                    err,
                    type(err),
                    number,
                )
                raise

        if "." in number:
            _, decimal = number.split(sep=".")
            return len(decimal)
        else:
            return 0



from typing import List

from .extensions import Extensions
from .gpx_element import GpxElement
from .way_point import WayPoint


class TrackSegment(GpxElement):
    """
    trksegType in GPX file.
    """

    fields = ["trkpt", "extensions"]
    mandatory_fields = []

    def __init__(
        self,
        tag: str = "trkseg",
        trkpt: List[WayPoint] = None,
        extensions: Extensions = None,
    ) -> None:
        """
        Initialize TrackSegment instance.

        Parameters
        ----------
        tag : str, optional
            XML tag, by default "trkseg"
        trkpt : List[WayPoint], optional
            List of track points, by default None
        extensions : Extensions, optional
            Extensions, by default None
        """
        self.tag: str = tag
        self.trkpt: List[WayPoint] = [] if trkpt is None else trkpt
        self.extensions: Extensions = extensions



import os
import warnings
from math import isclose
from typing import Optional, Tuple

import matplotlib
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.basemap import Basemap

from .plotter import Plotter


class MatplotlibAnimPlotter(Plotter):
    """
    GPX animated plotter based on Matplotlib.
    """

    def plot(
        self,
        figsize: Tuple[int, int] = (16, 9),
        size: float = 5,
        color: str = "#FFA800",
        background: Optional[str] = None,
        offset_percentage: float = 0.04,
        dpi: int = 96,
        interval: float = 20,
        fps: int = 24,
        bitrate: int = 1800,
        repeat: bool = True,
        title: Optional[str] = None,
        title_fontsize: int = 20,
        watermark: bool = False,
        file_path: str = None,
    ):
        """
        Plot (animation) GPX using Matplotlib.

        Crashes may be due to parametres exceeding system capabilities.
        Try reducing fps and/or bitrate.

        Args:
            figsize (Tuple[int, int], optional): Width and height of
                the plot. Defaults to (16, 9).
            size (float, optional): Size of the track. Defaults to 5.
            color (str, optional): Color of the track. Defaults to
                "#FFA800".
            background (Optional[str], optional): Map tiles to use.
                Possible choice are: None, "bluemarble",
                "shadedrelief", "etopo", "World_Imagery", "wms" or any
                server supported by
                `mpl_toolkits.basemap.Basemap.arcgisimage`. Defaults to
                None.
            offset_percentage (float, optional): Offset percentage to
                apply to the track bounding box. Defaults to 0.04.
            dpi (int, optional): Resolution of the animation. Defaults
                to 96.
            interval (float, optional): Interval between frames of the
                animation. Defaults to 20.
            fps (int, optional): FPS of the animation. Defaults to 24.
            bitrate (int, optional): Bit-rate of the animation.
                Defaults to 1800.
            repeat (bool, optional): Repeat the animation when viewed.
                Defaults to True.
            title (Optional[str], optional): Title of the plot.
                Defaults to None.
            title_fontsize (int, optional): Font size of the title of
                the plot. Defaults to 20.
            watermark (bool, optional): Watermark. Defaults to False.
            file_path (str, optional): Path to save the plot. Defaults
                to None.

        Raises:
            FileNotFoundError: Provided path does not exist.

        Returns:
            matplotlib.Figure: Animated plot of the GPX.
        """
        # Create dataframe containing data from the GPX file
        self._dataframe = self._gpx.to_pandas()

        # Retrieve useful data
        lat = self._dataframe["lat"].values
        lon = self._dataframe["lon"].values

        # Create figure
        fig = plt.figure(figsize=figsize)
        ax = fig.gca()

        # Compute track boundaries
        min_lat, min_lon, max_lat, max_lon = self._gpx.bounds()

        # Compute default offset
        delta_lat = abs(max_lat - min_lat)
        delta_lon = abs(max_lon - min_lon)
        delta_max = max(delta_lat, delta_lon)
        offset = delta_max * offset_percentage

        # Add default offset
        min_lat = max(-90, min_lat - offset)
        max_lat = min(max_lat + offset, 90)
        min_lon = max(-180, min_lon - offset)
        max_lon = min(max_lon + offset, 180)

        # Update min/max lat/lon to achieve correct aspect ratio
        lat_offset = 1e-5
        lon_offset = 1e-5
        delta_lat = abs(max_lat - min_lat)
        delta_lon = abs(max_lon - min_lon)
        r = delta_lon / delta_lat  # Current map aspect ratio
        r_ref = figsize[0] / figsize[1]  # Target map aspect ratio
        tolerance = 1e-3
        while not isclose(r, r_ref, abs_tol=tolerance):
            if r > r_ref:
                min_lat = max(-90, min_lat - lat_offset)
                max_lat = min(max_lat + lat_offset, 90)
                delta_lat = max_lat - min_lat
            if r < r_ref:
                min_lon = max(-180, min_lon - lon_offset)
                max_lon = min(max_lon + lon_offset, 180)
                delta_lon = max_lon - min_lon
            r = delta_lon / delta_lat

        # Create map
        map_ = Basemap(
            projection="cyl",
            llcrnrlon=min_lon,
            llcrnrlat=min_lat,
            urcrnrlon=max_lon,
            urcrnrlat=max_lat,
            ax=ax,
        )

        # Add background
        if background is None:
            pass
        elif background == "bluemarble":
            map_.bluemarble()
        elif background == "shadedrelief":
            map_.shadedrelief()
        elif background == "etopo":
            map_.etopo()
        elif background == "World_Imagery":
            map_.arcgisimage(service=background, dpi=dpi)
        elif background == "wms":
            wms_server = "http://www.ga.gov.au/gis/services/topography/Australian_Topography/MapServer/WMSServer"
            wms_server = "http://wms.geosignal.fr/metropole?"
            map_.wmsimage(
                wms_server, layers=["Communes", "Nationales", "Regions"], verbose=True
            )
        else:
            map_.arcgisimage(service=background, dpi=dpi, verbose=True)

        # Create empty line
        (line,) = fig.gca().plot([], [], color=color, linewidth=size)

        # Animation function
        def animate(i):
            # Clear line at the beginning of the animation
            if i == 0:
                line.set_xdata([])
                line.set_ydata([])
            try:
                line.set_xdata(np.concatenate((line.get_xdata(), np.array([lon[i]]))))
                line.set_ydata(np.concatenate((line.get_ydata(), np.array([lat[i]]))))
            except IndexError:
                warnings.warn("No more data point, you need to stop!!")

            return line

        # Create animation
        ani = animation.FuncAnimation(
            fig=fig,
            func=animate,
            frames=len(lat),
            interval=interval,  # Delay between frames in ms
            blit=False,
            repeat=repeat if file_path is None else False,
        )

        # Add title
        if title is not None:
            if watermark:
                fig.suptitle(title + "\n[made with ezGPX]", fontsize=title_fontsize)
            else:
                fig.suptitle(title, fontsize=title_fontsize)

        # Set figure layout
        fig.tight_layout()

        # Save plot
        if file_path is not None:
            # Check if provided path exists
            directory_path = os.path.dirname(os.path.realpath(file_path))
            if not os.path.exists(directory_path):
                raise FileNotFoundError("Provided path does not exist")
            writer = None
            if file_path.endswith(".mp4"):
                writer = animation.FFMpegWriter(
                    fps=fps, metadata={"artist": "ezGPX"}, bitrate=bitrate
                )
            elif file_path.endswith(".gif"):
                writer = animation.PillowWriter(
                    fps=fps, metadata={"artist": "ezGPX"}, bitrate=bitrate
                )
            ani.save(file_path, writer=writer)

        return fig



from .bounds import *
from .copyright import *
from .email import *
from .extensions import *
from .gpx import *
from .gpx_element import *
from .link import *
from .metadata import *
from .person import *
from .point import *
from .point_segment import *
from .route import *
from .track import *
from .track_segment import *
from .way_point import *
